from typing import Any, Union, Sequence
import numpy as np

try:
    # New in Python-3.11
    from typing_extension import Self
except ImportError:
    Self = Any  # type: ignore

try:
    # New in Python-3.8
    from typing import Literal
except ImportError:
    class _Literal:
        def __getitem__(self, index):
            return Any
    Literal = _Literal()  # type: ignore

try:
    # Needs numpy-1.20:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    ArrayLike = Any  # type: ignore
    DTypeLike = Any  # type: ignore

ArrayLike1D = ArrayLike
ArrayLike2D = ArrayLike

ArrayND = np.ndarray
Array1D = ArrayND
Array2D = ArrayND
Array3D = ArrayND
Array4D = ArrayND

# Used for sequences of three numbers:
Vector = Union[Sequence[float], Array1D]
IntVector = Union[Sequence[int], Array1D]



from ase import Atom
from ase.units import Ha
from ase.parallel import parprint

from gpaw import GPAW

from gpaw.cluster import Cluster
from gpaw.pes.state import BoundState, H1s
from gpaw.pes.ds_beta import CrossSectionBeta


def test_ds_beta(in_tmp_dir):
    ngauss = 2

    h = .3
    box = 3.

    gpwname = 'H1s.gpw'
    if 1:
        c = GPAW(mode='fd', xc='PBE', nbands=-1, h=h)
        s = Cluster([Atom('H')])
        s.minimal_box(box, h=h)
        c.calculate(s)
        c.write(gpwname, 'all')
    else:
        c = GPAW(gpwname)
        s = c.get_atoms()
        c.converge_wave_functions()
    cm = s.get_center_of_mass()
    Ekin = 1.

    for form, title in [('L', 'length form'),
                        ('V', 'velocity form')]:
        parprint('--', title)
        ds = []
        for analytic in [True, False]:
            if analytic:
                initial = H1s(c.density.gd, cm)
            else:
                initial = BoundState(c, 0, 0)
                initial.set_energy(- Ha / 2.)

            csb = CrossSectionBeta(initial=initial,
                                   final=None,
                                   r0=cm, ngauss=ngauss, form=form)
            if analytic:
                ds.append(initial.get_ds(Ekin, form))
                parprint('analytic 1s energy, beta, ds %5.3f' %
                         (Ekin + Ha / 2.), end='')
                parprint('%8.4f %12.5f' % (2, ds[-1]))
            ds.append(csb.get_ds(Ekin))
            parprint('numeric  1s energy, beta, ds %5.3f' %
                     (Ekin + Ha / 2.), end='')
            parprint('%8.4f %12.5f' % (csb.get_beta(Ekin), ds[-1]))
        parprint('error analytic GS:',
                 int(100 * abs(ds[1] / ds[0] - 1.) + .5), '%')
        assert abs(ds[1] / ds[0] - 1.) < 0.31
        parprint('error numeric GS:',
                 int(100 * abs(ds[2] / ds[0] - 1.) + .5), '%')
        assert abs(ds[2] / ds[0] - 1.) < 0.2



from gpaw.analyse.observers import Observer


class TDDFTObserver(Observer):

    def __init__(self, paw, interval):
        Observer.__init__(self, interval)
        self.timer = paw.timer
        if hasattr(paw, 'time') and hasattr(paw, 'niter'):
            paw.attach(self, interval, paw)

    def update(self, paw):
        self.timer.start('%s update' % self.__class__.__name__)
        self._update(paw)
        self.timer.stop('%s update' % self.__class__.__name__)

    def _update(self, paw):
        raise NotImplementedError()

    def write_restart(self):
        """Write restart file.

        Optional method that will be called by RestartFileWriter
        when writing the calculator restart file.
        """

    def __str__(self):
        return self.__class__.__name__



"""Module for evaluating two-center integrals.

Contains classes for evaluating integrals of the form::

             /
            |   _   _a    _   _b   _
    Theta = | f(r - R ) g(r - R ) dr ,
            |
           /

with f and g each being given as a radial function times a spherical
harmonic.

Important classes
-----------------

Low-level:

 * OverlapExpansion: evaluate the overlap between a pair of functions (or a
   function with itself) for some displacement vector: <f | g>.  An overlap
   expansion can be created once for a pair of splines f and g, and actual
   values of the overlap can then be evaluated for several different
   displacement vectors.
 * FourierTransformer: create OverlapExpansion object from pair of splines.

Mid-level:

 * TwoSiteOverlapExpansions: evaluate overlaps between two *sets* of functions,
   where functions in the same set reside on the same location: <f_j | g_j>.
 * TwoSiteOverlapCalculator: create TwoSiteOverlapExpansions object from
   pair of lists of splines.

High-level:

 * ManySiteOverlapExpansions:  evaluate overlaps with many functions in many
   locations: <f_aj | g_aj>.
 * ManySiteOverlapCalculator: create ManySiteOverlapExpansions object from
   pair of lists of splines nested by atom and orbital number.

The low-level classes do the actual work, while the higher-level ones depend
on the lower-level ones.

"""

from math import pi

import numpy as np

from ase.neighborlist import PrimitiveNeighborList
from ase.data import covalent_radii
from ase.units import Bohr

import _gpaw
from gpaw.ffbt import ffbt, FourierBesselTransformer
from gpaw.gaunt import gaunt
from gpaw.spherical_harmonics import Yl, nablarlYL
from gpaw.spline import Spline
from gpaw.utilities.tools import tri2full
from gpaw.utilities.timing import nulltimer

timer = nulltimer  # XXX global timer object, only for hacking

UL = 'L'


class BaseOverlapExpansionSet:
    def __init__(self, shape):
        self.shape = shape

    def zeros(self, shape=(), dtype=float):
        return np.zeros(shape + self.shape, dtype=dtype)


class OverlapExpansion(BaseOverlapExpansionSet):
    """A list of real-space splines representing an overlap integral."""
    def __init__(self, la, lb, spline_l):
        self.la = la
        self.lb = lb
        self.lmaxgaunt = max(la, lb)
        self.spline_l = spline_l
        self.lmaxspline = (la + lb) % 2 + 2 * len(self.spline_l)
        BaseOverlapExpansionSet.__init__(self, (2 * la + 1, 2 * lb + 1))
        self.cspline_l = [spline.spline for spline in self.spline_l]

    def evaluate(self, r, rlY_lm, G_LLL, x_mi, _nil=np.empty(0)):
        _gpaw.tci_overlap(self.la, self.lb, G_LLL, self.cspline_l,
                          r, rlY_lm, x_mi,
                          False, _nil, _nil, _nil)

    def derivative(self, r, Rhat_c, rlY_L, G_LLL, drlYdR_Lc, dxdR_cmi,
                   _nil=np.empty(0)):
        # timer.start('deriv')
        _gpaw.tci_overlap(self.la, self.lb, G_LLL, self.cspline_l,
                          r, rlY_L, _nil,
                          True, Rhat_c, drlYdR_Lc, dxdR_cmi)
        # timer.stop('deriv')


class TwoSiteOverlapExpansions(BaseOverlapExpansionSet):
    def __init__(self, la_j, lb_j, oe_jj):
        self.oe_jj = oe_jj
        shape = (sum([2 * l + 1 for l in la_j]),
                 sum([2 * l + 1 for l in lb_j]))
        BaseOverlapExpansionSet.__init__(self, shape)
        if oe_jj.size == 0:
            self.lmaxgaunt = 0
            self.lmaxspline = 0
        else:
            self.lmaxgaunt = max(oe.lmaxgaunt for oe in oe_jj.ravel())
            self.lmaxspline = max(oe.lmaxspline for oe in oe_jj.ravel())

    def slice(self, x_xMM):
        assert x_xMM.shape[-2:] == self.shape
        Ma1 = 0
        for j1, oe_j in enumerate(self.oe_jj):
            Mb1 = 0
            Ma2 = Ma1
            for j2, oe in enumerate(oe_j):
                Ma2 = Ma1 + oe.shape[0]
                Mb2 = Mb1 + oe.shape[1]
                yield x_xMM[..., Ma1:Ma2, Mb1:Mb2], oe
                Mb1 = Mb2
            Ma1 = Ma2

    def evaluate(self, r, rlY_lm):
        timer.start('tsoe eval')
        x_MM = self.zeros()
        G_LLL = gaunt(self.lmaxgaunt)
        rlY_L = rlY_lm.toarray(self.lmaxspline)
        for x_mm, oe in self.slice(x_MM):
            oe.evaluate(r, rlY_L, G_LLL, x_mm)
        timer.stop('tsoe eval')
        return x_MM

    def derivative(self, r, Rhat, rlY_lm, drlYdR_lmc):
        x_cMM = self.zeros((3,))
        G_LLL = gaunt(self.lmaxgaunt)
        rlY_L = rlY_lm.toarray(self.lmaxspline)
        drlYdR_Lc = drlYdR_lmc.toarray(self.lmaxspline)
        # print(drlYdR_lmc.shape)
        for x_cmm, oe in self.slice(x_cMM):
            oe.derivative(r, Rhat, rlY_L, G_LLL, drlYdR_Lc, x_cmm)
        return x_cMM


class ManySiteOverlapExpansions(BaseOverlapExpansionSet):
    def __init__(self, tsoe_II, I1_a, I2_a):
        self.tsoe_II = tsoe_II
        self.I1_a = I1_a
        self.I2_a = I2_a

        M1 = 0
        M1_a = []
        for I in I1_a:
            M1_a.append(M1)
            M1 += tsoe_II[I, 0].shape[0]
        self.M1_a = M1_a

        M2 = 0
        M2_a = []
        for I in I2_a:
            M2_a.append(M2)
            M2 += tsoe_II[0, I].shape[1]
        self.M2_a = M2_a

        shape = (sum([tsoe_II[I, 0].shape[0] for I in I1_a]),
                 sum([tsoe_II[0, I].shape[1] for I in I2_a]))
        assert (M1, M2) == shape
        BaseOverlapExpansionSet.__init__(self, shape)

    def get(self, a1, a2):
        return self.tsoe_II[self.I1_a[a1], self.I2_a[a2]]

    def getslice(self, a1, a2, x_xMM):
        I1 = self.I1_a[a1]
        I2 = self.I2_a[a2]
        tsoe = self.tsoe_II[I1, I2]
        Mstart1 = self.M1_a[a1]
        Mend1 = Mstart1 + tsoe.shape[0]
        Mstart2 = self.M2_a[a2]
        Mend2 = Mstart2 + tsoe.shape[1]
        return x_xMM[..., Mstart1:Mend1, Mstart2:Mend2], tsoe

    def evaluate_slice(self, disp, x_qxMM):
        x_qxmm, oe = self.getslice(disp.a1, disp.a2, x_qxMM)
        disp.evaluate_overlap(oe, x_qxmm)


class DomainDecomposedExpansions(BaseOverlapExpansionSet):
    def __init__(self, msoe, local_indices):
        self.msoe = msoe
        self.local_indices = local_indices
        BaseOverlapExpansionSet.__init__(self, msoe.shape)

    def evaluate_slice(self, disp, x_xqMM):
        if disp.a2 in self.local_indices:
            self.msoe.evaluate_slice(disp, x_xqMM)


class ManySiteDictionaryWrapper(DomainDecomposedExpansions):
    # Used with dictionaries such as P_aqMi and dPdR_aqcMi

    def getslice(self, a1, a2, xdict_aqxMi):
        msoe = self.msoe
        tsoe = msoe.tsoe_II[msoe.I1_a[a1], msoe.I2_a[a2]]
        Mstart = self.msoe.M1_a[a1]
        Mend = Mstart + tsoe.shape[0]
        return xdict_aqxMi[a2][..., Mstart:Mend, :], tsoe

    def evaluate_slice(self, disp, x_aqxMi):
        if disp.a2 in x_aqxMi:
            x_qxmi, oe = self.getslice(disp.a1, disp.a2, x_aqxMi)
            disp.evaluate_overlap(oe, x_qxmi)
        if disp.a1 in x_aqxMi and (disp.a1 != disp.a2):
            x2_qxmi, oe2 = self.getslice(disp.a2, disp.a1, x_aqxMi)
            rdisp = disp.reverse()
            rdisp.evaluate_overlap(oe2, x2_qxmi)


class BlacsOverlapExpansions(BaseOverlapExpansionSet):
    def __init__(self, msoe, local_indices, Mmystart, mynao):
        self.msoe = msoe
        self.local_indices = local_indices
        BaseOverlapExpansionSet.__init__(self, msoe.shape)

        self.Mmystart = Mmystart
        self.mynao = mynao

        M_a = msoe.M1_a
        natoms = len(M_a)
        a = 0
        while a < natoms and M_a[a] <= Mmystart:
            a += 1
        a -= 1
        self.astart = a

        while a < natoms and M_a[a] < Mmystart + mynao:
            a += 1
        self.aend = a

    def evaluate_slice(self, disp, x_xqNM):
        a1 = disp.a1
        a2 = disp.a2
        if a2 in self.local_indices and (self.astart <= a1 < self.aend):
            # assert a2 <= a1
            msoe = self.msoe
            I1 = msoe.I1_a[a1]
            I2 = msoe.I2_a[a2]
            tsoe = msoe.tsoe_II[I1, I2]
            x_qxmm = tsoe.zeros(x_xqNM.shape[:-2], dtype=x_xqNM.dtype)
            disp.evaluate_overlap(tsoe, x_qxmm)
            Mstart1 = msoe.M1_a[a1] - self.Mmystart
            Mend1 = Mstart1 + tsoe.shape[0]
            Mstart1b = max(0, Mstart1)
            Mend1b = min(self.mynao, Mend1)
            Mstart2 = msoe.M2_a[a2]
            Mend2 = Mstart2 + tsoe.shape[1]
            x_xqNM[..., Mstart1b:Mend1b, Mstart2:Mend2] += \
                x_qxmm[..., Mstart1b - Mstart1:Mend1b - Mstart1, :]
        # This is all right as long as we are only interested in a2 <= a1
        # if a1 in self.local_indices and a2 < a1 and (self.astart <=
        #                                              a2 < self.aend):
        #     self.evaluate_slice(disp.reverse(), x_xqNM)


class NeighborPairs:
    """Class for looping over pairs of atoms using a neighbor list."""
    def __init__(self, cutoff_a, cell_cv, pbc_c, self_interaction):
        self.neighbors = PrimitiveNeighborList(
            cutoff_a, skin=0, sorted=True,
            self_interaction=self_interaction,
            use_scaled_positions=True)
        self.cell_cv = cell_cv
        self.pbc_c = pbc_c

    def set_positions(self, spos_ac):
        self.spos_ac = spos_ac
        self.neighbors.update(self.pbc_c, self.cell_cv, spos_ac)

    def iter(self):
        cell_cv = self.cell_cv
        for a1, spos1_c in enumerate(self.spos_ac):
            a2_a, offsets = self.neighbors.get_neighbors(a1)
            for a2, offset in zip(a2_a, offsets):
                spos2_c = self.spos_ac[a2] + offset
                R_c = np.dot(spos2_c - spos1_c, cell_cv)
                yield a1, a2, R_c, offset


class PairFilter:
    def __init__(self, pairs):
        self.pairs = pairs

    def set_positions(self, spos_ac):
        self.pairs.set_positions(spos_ac)

    def iter(self):
        return self.pairs.iter()


class PairsWithSelfinteraction(PairFilter):
    def iter(self):
        for a1, a2, R_c, offset in self.pairs.iter():
            yield a1, a2, R_c, offset
            if a1 == a2 and offset.any():
                yield a1, a1, -R_c, -offset


class OppositeDirection(PairFilter):
    def iter(self):
        for a1, a2, R_c, offset in self.pairs.iter():
            yield a2, a1, -R_c, -offset


class FourierTransformer(FourierBesselTransformer):
    def calculate_overlap_expansion(self, phit1phit2_q, l1, l2):
        """Calculate list of splines for one overlap integral.

        Given two Fourier transformed functions, return list of splines
        in real space necessary to evaluate their overlap.

          phi  (q) * phi  (q) --> [phi    (r), ..., phi    (r)] .
             l1         l2            lmin             lmax

        The overlap <phi1 | phi2> can then be calculated by linear
        combinations of the returned splines with appropriate Gaunt
        coefficients.
        """
        lmax = l1 + l2
        splines = []
        R = np.arange(self.Q) * self.dr
        R1 = R.copy()
        R1[0] = 1.0
        k1 = self.k_q.copy()
        k1[0] = 1.0
        a_q = phit1phit2_q
        for l in range(lmax % 2, lmax + 1, 2):
            a_g = (8 * ffbt(l, a_q * k1**(-2 - lmax - l), self.k_q, R) /
                   R1**(2 * l + 1))
            if l == 0:
                a_g[0] = 8 * np.sum(a_q * k1**(-lmax)) * self.dk
            else:
                a_g[0] = a_g[1]  # XXXX
            a_g *= (-1)**((l1 - l2 - l) // 2)
            n = len(a_g) // 256
            s = Spline(l, 2 * self.rcut, np.concatenate((a_g[::n], [0.0])))
            splines.append(s)
        return OverlapExpansion(l1, l2, splines)

    def laplacian(self, f_jq):
        return 0.5 * f_jq * self.k_q**2.0


class TwoSiteOverlapCalculator:
    def __init__(self, transformer):
        self.transformer = transformer

    def transform(self, f_j):
        f_jq = np.array([self.transformer.transform(f) for f in f_j])
        return f_jq

    def calculate_expansions(self, la_j, fa_jq, lb_j, fb_jq):
        nja = len(la_j)
        njb = len(lb_j)
        assert nja == len(fa_jq) and njb == len(fb_jq)
        oe_jj = np.empty((nja, njb), dtype=object)
        for ja, (la, fa_q) in enumerate(zip(la_j, fa_jq)):
            for jb, (lb, fb_q) in enumerate(zip(lb_j, fb_jq)):
                a_q = fa_q * fb_q
                oe = self.transformer.calculate_overlap_expansion(a_q, la, lb)
                oe_jj[ja, jb] = oe
        return TwoSiteOverlapExpansions(la_j, lb_j, oe_jj)

    def calculate_kinetic_expansions(self, l_j, f_jq):
        t_jq = self.transformer.laplacian(f_jq)
        return self.calculate_expansions(l_j, f_jq, l_j, t_jq)

    def laplacian(self, f_jq):
        t_jq = self.transformer.laplacian(f_jq)
        return t_jq


class ManySiteOverlapCalculator:
    def __init__(self, twosite_calculator, I1_a, I2_a):
        """Create VeryManyOverlaps object.

        twosite_calculator: instance of TwoSiteOverlapCalculator
        I_a: mapping from atom index (as in spos_a) to unique atom type"""
        self.twosite_calculator = twosite_calculator
        self.I1_a = I1_a
        self.I2_a = I2_a

    def transform(self, f_Ij):
        f_Ijq = [self.twosite_calculator.transform(f_j) for f_j in f_Ij]
        return f_Ijq

    def calculate_expansions(self, l1_Ij, f1_Ijq, l2_Ij, f2_Ijq):
        # Naive implementation, just loop over everything
        # We should only need half of them
        nI1 = len(l1_Ij)
        nI2 = len(l2_Ij)
        assert len(l1_Ij) == len(f1_Ijq) and len(l2_Ij) == len(f2_Ijq)
        tsoe_II = np.empty((nI1, nI2), dtype=object)
        calc = self.twosite_calculator
        for I1, (l1_j, f1_jq) in enumerate(zip(l1_Ij, f1_Ijq)):
            for I2, (l2_j, f2_jq) in enumerate(zip(l2_Ij, f2_Ijq)):
                tsoe = calc.calculate_expansions(l1_j, f1_jq, l2_j, f2_jq)
                tsoe_II[I1, I2] = tsoe
        return ManySiteOverlapExpansions(tsoe_II, self.I1_a, self.I2_a)

    def calculate_kinetic_expansions(self, l_Ij, f_Ijq):
        t_Ijq = [self.twosite_calculator.laplacian(f_jq) for f_jq in f_Ijq]
        return self.calculate_expansions(l_Ij, f_Ijq, l_Ij, t_Ijq)


class AtomicDisplacement:
    def __init__(self, factory, a1, a2, R_c, offset, phases):
        self.factory = factory
        self.a1 = a1
        self.a2 = a2
        self.R_c = R_c
        self.offset = offset
        self.phases = phases
        self.r = np.linalg.norm(R_c)
        self._set_spherical_harmonics(R_c)

    def _set_spherical_harmonics(self, R_c):
        self.rlY_lm = LazySphericalHarmonics(R_c)

    # XXX new
    def evaluate_direct(self, oe, dst_xqmm):
        src_xmm = self.evaluate_direct_without_phases(oe)
        self.phases.apply(src_xmm, dst_xqmm)

    # XXX new
    def evaluate_direct_without_phases(self, oe):
        return oe.evaluate(self.r, self.rlY_lm)

    # XXX clean up unnecessary methods
    def overlap_without_phases(self, oe):
        return oe.evaluate(self.r, self.rlY_lm)

    def _evaluate_without_phases(self, oe):
        return self.overlap_without_phases(oe)

    def evaluate_overlap(self, oe, dst_xqmm):
        src_xmm = self._evaluate_without_phases(oe)
        timer.start('phases')
        self.phases.apply(src_xmm, dst_xqmm)
        timer.stop('phases')

    def reverse(self):
        return self.factory.displacementclass(self.factory, self.a2, self.a1,
                                              -self.R_c, -self.offset,
                                              self.phases.inverse())


class LazySphericalHarmonics:
    """Class for caching spherical harmonics as they are calculated.

    Behaves like a list Y_lm, but really calculates (or retrieves) Y_m
    once a given value of l is __getitem__'d."""
    def __init__(self, R_c):
        self.R_c = np.asarray(R_c)
        self.Y_lm = {}
        self.Y_lm1 = np.empty(0)
        # self.dYdr_lmc = {}
        self.lmax = 0

    def evaluate(self, l):
        rlY_m = np.empty(2 * l + 1)
        Yl(l, self.R_c, rlY_m)
        return rlY_m

    def __getitem__(self, l):
        Y_m = self.Y_lm.get(l)
        if Y_m is None:
            Y_m = self.evaluate(l)
            self.Y_lm[l] = Y_m
        return Y_m

    def toarray(self, lmax):
        if lmax > self.lmax:
            self.Y_lm1 = np.concatenate([self.Y_lm1] +
                                        [self.evaluate(l).ravel()
                                         for l in range(self.lmax, lmax)])
            self.lmax = lmax
        return self.Y_lm1


class LazySphericalHarmonicsDerivative(LazySphericalHarmonics):
    def evaluate(self, l):
        drlYdR_mc = np.empty((2 * l + 1, 3))
        L0 = l**2
        for m in range(2 * l + 1):
            drlYdR_mc[m, :] = nablarlYL(L0 + m, self.R_c)
        return drlYdR_mc


class DerivativeAtomicDisplacement(AtomicDisplacement):
    def _set_spherical_harmonics(self, R_c):
        self.rlY_lm = LazySphericalHarmonics(R_c)
        self.drlYdr_lmc = LazySphericalHarmonicsDerivative(R_c)

        if R_c.any():
            self.Rhat_c = R_c / self.r
        else:
            self.Rhat_c = np.zeros(3)

    def derivative_without_phases(self, oe):
        return oe.derivative(self.r, self.Rhat_c, self.rlY_lm, self.drlYdr_lmc)

    def _evaluate_without_phases(self, oe):  # override
        return self.derivative_without_phases(oe)


class NullPhases:
    def __init__(self, ibzk_qc, offset):
        pass

    def apply(self, src_xMM, dst_qxMM):
        assert len(dst_qxMM) == 1
        dst_qxMM[0][:] += src_xMM

    def inverse(self):
        return self


class BlochPhases:
    def __init__(self, ibzk_qc, offset):
        self.phase_q = np.exp(-2j * pi * np.dot(ibzk_qc, offset))
        self.ibzk_qc = ibzk_qc
        self.offset = offset

    def apply(self, src_xMM, dst_qxMM):
        assert dst_qxMM.dtype == complex, dst_qxMM.dtype
        for phase, dst_xMM in zip(self.phase_q, dst_qxMM):
            dst_xMM[:] += phase * src_xMM

    def inverse(self):
        return BlochPhases(-self.ibzk_qc, self.offset)


class TwoCenterIntegralCalculator:
    # This class knows how to apply phases, and whether to call the
    # various derivative() or evaluate() methods
    def __init__(self, ibzk_qc=None, derivative=False):
        if derivative:
            displacementclass = DerivativeAtomicDisplacement
        else:
            displacementclass = AtomicDisplacement
        self.displacementclass = displacementclass

        if ibzk_qc is None or not ibzk_qc.any():
            self.phaseclass = NullPhases
        else:
            self.phaseclass = BlochPhases
        self.ibzk_qc = ibzk_qc
        self.derivative = derivative

    def calculate(self, atompairs, expansions, arrays):
        for disp in self.iter(atompairs):
            for expansion, X_qxMM in zip(expansions, arrays):
                expansion.evaluate_slice(disp, X_qxMM)

    def iter(self, atompairs):
        for a1, a2, R_c, offset in atompairs.iter():
            # if a1 == a2 and self.derivative:
            #     continue
            phase_applier = self.phaseclass(self.ibzk_qc, offset)
            yield self.displacementclass(self, a1, a2, R_c, offset,
                                         phase_applier)


class NewTwoCenterIntegrals:
    def __init__(self, cell_cv, pbc_c, setups, ibzk_qc, gamma):
        self.cell_cv = cell_cv
        self.pbc_c = pbc_c
        self.ibzk_qc = ibzk_qc
        self.gamma = gamma

        timer.start('tci init')
        cutoff_I = []
        setups_I = setups.setups.values()
        I_setup = {}
        for I, setup in enumerate(setups_I):
            I_setup[setup] = I
            cutoff_I.append(max([func.get_cutoff()
                                 for func
                                 in setup.basis_functions_J + setup.pt_j]))

        I_a = []
        for setup in setups:
            I_a.append(I_setup[setup])

        cutoff_a = [cutoff_I[I] for I in I_a]

        self.cutoff_a = cutoff_a  # convenient for writing the new new overlap
        self.I_a = I_a
        self.setups_I = setups_I
        self.atompairs = PairsWithSelfinteraction(NeighborPairs(cutoff_a,
                                                                cell_cv,
                                                                pbc_c,
                                                                True))

        scale = 0.01  # XXX minimal distance scale
        cutoff_close_a = [covalent_radii[int(s.Z)] / Bohr * scale
                          for s in setups]
        self.atoms_close = NeighborPairs(cutoff_close_a, cell_cv, pbc_c, False)

        rcut = max(cutoff_I + [0.001])
        transformer = FourierTransformer(rcut, N=2**10)
        tsoc = TwoSiteOverlapCalculator(transformer)
        self.msoc = ManySiteOverlapCalculator(tsoc, I_a, I_a)
        self.calculate_expansions()

        self.calculate = self.evaluate  # XXX compatibility

        self.set_matrix_distribution(None, None)
        timer.stop('tci init')

    def set_matrix_distribution(self, Mmystart, mynao):
        """Distribute matrices using BLACS."""
        # Range of basis functions for BLACS distribution of matrices:
        self.Mmystart = Mmystart
        self.mynao = mynao
        self.blacs = mynao is not None

    def calculate_expansions(self):
        timer.start('tci calc exp')
        phit_Ij = [setup.basis_functions_J for setup in self.setups_I]
        l_Ij = []
        for phit_j in phit_Ij:
            l_Ij.append([phit.get_angular_momentum_number()
                         for phit in phit_j])

        pt_l_Ij = [setup.l_j for setup in self.setups_I]
        pt_Ij = [setup.pt_j for setup in self.setups_I]
        phit_Ijq = self.msoc.transform(phit_Ij)
        pt_Ijq = self.msoc.transform(pt_Ij)

        msoc = self.msoc

        self.Theta_expansions = msoc.calculate_expansions(l_Ij, phit_Ijq,
                                                          l_Ij, phit_Ijq)
        self.T_expansions = msoc.calculate_kinetic_expansions(l_Ij, phit_Ijq)
        self.P_expansions = msoc.calculate_expansions(l_Ij, phit_Ijq,
                                                      pt_l_Ij, pt_Ijq)
        timer.stop('tci calc exp')

    def _calculate(self, calc, spos_ac, Theta_qxMM, T_qxMM, P_aqxMi):
        Theta_qxMM.fill(0.0)
        T_qxMM.fill(0.0)
        for P_qxMi in P_aqxMi.values():
            P_qxMi.fill(0.0)

        if 1:  # XXX
            self.atoms_close.set_positions(spos_ac)
            wrk = [x for x in self.atoms_close.iter()]
            if len(wrk) != 0:
                txt = ''
                for a1, a2, R_c, offset in wrk:
                    txt += 'Atom %d and Atom %d in cell (%d, %d, %d)\n' % \
                        (a1, a2, offset[0], offset[1], offset[2])
                raise RuntimeError('Atoms too close!\n' + txt)

        self.atompairs.set_positions(spos_ac)

        if self.blacs:
            # S and T matrices are distributed:
            expansions = [
                BlacsOverlapExpansions(self.Theta_expansions,
                                       P_aqxMi, self.Mmystart, self.mynao),
                BlacsOverlapExpansions(self.T_expansions,
                                       P_aqxMi, self.Mmystart, self.mynao)]
        else:
            expansions = [DomainDecomposedExpansions(self.Theta_expansions,
                                                     P_aqxMi),
                          DomainDecomposedExpansions(self.T_expansions,
                                                     P_aqxMi)]

        expansions.append(ManySiteDictionaryWrapper(self.P_expansions,
                                                    P_aqxMi))
        arrays = [Theta_qxMM, T_qxMM, P_aqxMi]
        timer.start('tci calculate')
        calc.calculate(OppositeDirection(self.atompairs), expansions, arrays)
        timer.stop('tci calculate')

    def evaluate(self, spos_ac, Theta_qMM, T_qMM, P_aqMi):
        calc = TwoCenterIntegralCalculator(self.ibzk_qc, derivative=False)
        self._calculate(calc, spos_ac, Theta_qMM, T_qMM, P_aqMi)
        if not self.blacs:
            for X_MM in list(Theta_qMM) + list(T_qMM):
                tri2full(X_MM, UL=UL)

    def derivative(self, spos_ac, dThetadR_qcMM, dTdR_qcMM, dPdR_aqcMi):
        calc = TwoCenterIntegralCalculator(self.ibzk_qc, derivative=True)
        self._calculate(calc, spos_ac, dThetadR_qcMM, dTdR_qcMM, dPdR_aqcMi)

        def antihermitian(src, dst):
            np.conj(-src, dst)

        if not self.blacs:
            for X_cMM in list(dThetadR_qcMM) + list(dTdR_qcMM):
                for X_MM in X_cMM:
                    tri2full(X_MM, UL=UL, map=antihermitian)

    calculate_derivative = derivative  # XXX compatibility

    def estimate_memory(self, mem):
        mem.subnode('TCI not impl.', 0)



"""Test selfconsistent RSF calculation with Yukawa potential including vc."""
import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW, KohnShamConvergenceError
from gpaw.xc.hybrid import HybridXC
from gpaw.poisson import PoissonSolver
from gpaw.occupations import FermiDirac
from gpaw.test import gen
from gpaw.eigensolvers import RMMDIIS
from gpaw.cluster import Cluster

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


def test_exx_exx_scf(in_tmp_dir, add_cwd_to_setup_paths):
    h = 0.3

    # No energies - simpley convergence test, esp. for 3d TM

    # for atom in ['F', 'Cl', 'Br', 'Cu', 'Ag']:
    for atom in ['Ti']:
        gen(atom, xcname='PBE', scalarrel=False, exx=True)
        work_atom = Cluster(Atoms(atom, [(0, 0, 0)]))
        work_atom.minimal_box(4, h=h)
        work_atom.translate([0.01, 0.02, 0.03])
        work_atom.set_initial_magnetic_moments([2.0])
        calculator = GPAW(mode='fd',
                          convergence={'energy': 0.01,
                                       'eigenstates': 3,
                                       'density': 3},
                          eigensolver=RMMDIIS(),
                          poissonsolver=PoissonSolver(use_charge_center=True),
                          occupations=FermiDirac(width=0.0, fixmagmom=True),
                          h=h, maxiter=35,  # Up to 24 are needed by now
                          xc=HybridXC('PBE0'),
                          txt=atom + '-PBE0.txt')
        work_atom.calc = calculator
        try:
            work_atom.get_potential_energy()
        except KohnShamConvergenceError:
            pass
        assert calculator.scf.converged, 'Calculation not converged'



import numpy as np
from ase.units import Hartree

import _gpaw
from gpaw.xc import XC
from gpaw.xc.kernel import XCKernel
from gpaw.xc.libxc import LibXC
from gpaw.xc.vdw import VDWFunctional
from gpaw import debug


class BEE2(XCKernel):
    """GGA exchange expanded in Legendre polynomials."""
    def __init__(self, parameters=None):
        """BEE2.

        parameters: array
            [transformation,0.0,[orders],[coefs]].

        """

        if parameters is None:
            parameters = (
                [4.0, 0.0] + list(range(30)) +
                [1.516501714304992365356, 0.441353209874497942611,
                 -0.091821352411060291887, -0.023527543314744041314,
                 0.034188284548603550816, 0.002411870075717384172,
                 -0.014163813515916020766, 0.000697589558149178113,
                 0.009859205136982565273, -0.006737855050935187551,
                 -0.001573330824338589097, 0.005036146253345903309,
                 -0.002569472452841069059, -0.000987495397608761146,
                 0.002033722894696920677, -0.000801871884834044583,
                 -0.000668807872347525591, 0.001030936331268264214,
                 -0.000367383865990214423, -0.000421363539352619543,
                 0.000576160799160517858, -0.000083465037349510408,
                 -0.000445844758523195788, 0.000460129009232047457,
                 -0.000005231775398304339, -0.000423957047149510404,
                 0.000375019067938866537, 0.000021149381251344578,
                 -0.000190491156503997170, 0.000073843624209823442])
        else:
            assert len(parameters) > 2
            assert np.mod(len(parameters), 2) == 0
            assert parameters[1] == 0.0

        parameters = np.array(parameters, dtype=float).ravel()
        self.xc = _gpaw.XCFunctional(17, parameters)
        self.type = 'GGA'
        self.name = 'BEE2'


class BEEVDWKernel(XCKernel):
    """Kernel for BEEVDW functionals."""
    def __init__(self, bee, xcoefs, ldac, ggac):
        """BEEVDW kernel.

        parameters:

        bee : str
            choose BEE1 or BEE2 exchange basis expansion.
        xcoefs : array
            coefficients for exchange.
        ldac : float
            coefficient for LDA correlation.
        pbec : float
            coefficient for PBE correlation.

        """

        if bee == 'BEE2':
            self.BEE = BEE2(xcoefs)
            self.GGAc = LibXC('GGA_C_PBE')
            self.xtype = 'GGA'
            self.type = 'GGA'
        elif bee == 'BEE3':
            self.BEE = LibXC('MGGA_X_MBEEFVDW')
            self.GGAc = LibXC('GGA_C_PBE_SOL')
            self.xtype = 'MGGA'
            self.type = 'MGGA'
        else:
            raise ValueError('Unknown BEE exchange: %s', bee)

        self.LDAc = LibXC('LDA_C_PW')
        self.ldac = ldac
        self.ggac = ggac
        if bee in ['BEE1', 'BEE2']:
            self.ggac -= 1.0
        self.name = 'BEEVDW'

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg=None, dedsigma_xg=None,
                  tau_sg=None, dedtau_sg=None):
        if debug:
            self.check_arguments(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                                 tau_sg, dedtau_sg)

        if self.xtype == 'GGA':
            self.BEE.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        elif self.xtype == 'MGGA':
            self.BEE.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                               tau_sg, dedtau_sg)
        else:
            raise ValueError('Unexpected value of xtype:', self.xtype)

        e0_g = np.empty_like(e_g)
        dedn0_sg = np.empty_like(dedn_sg)
        dedsigma0_xg = np.empty_like(dedsigma_xg)
        for coef, kernel in [(self.ldac, self.LDAc),
                             (self.ggac, self.GGAc)]:
            dedn0_sg[:] = 0.0
            kernel.calculate(e0_g, n_sg, dedn0_sg, sigma_xg, dedsigma0_xg)
            e_g += coef * e0_g
            dedn_sg += coef * dedn0_sg
            if kernel.type == 'GGA':
                dedsigma_xg += coef * dedsigma0_xg


class BEEFEnsemble:
    """BEEF ensemble error estimation."""
    def __init__(self, calc):
        """BEEF ensemble

        parameters:

        calc: object
            Calculator holding a selfconsistent BEEF type electron density.
            May be BEEF-vdW or mBEEF.
        """

        self.calc = calc

        self.e_dft = None
        self.e0 = None

        # determine functional and read parameters
        self.xc = self.calc.get_xc_functional()
        if self.xc == 'BEEF-vdW':
            self.bee_type = 1
        elif self.xc == 'mBEEF':
            self.bee_type = 2
            self.max_order = 8
            self.trans = [6.5124, -1.0]
            self.calc.converge_wave_functions()
            self.calc.log('wave functions converged')
        elif self.xc == 'mBEEF-vdW':
            self.bee_type = 3
            self.max_order = 5
            self.trans = [6.5124, -1.0]
            self.calc.converge_wave_functions()
            self.calc.log('wave functions converged')
        else:
            raise NotImplementedError('xc = %s not implemented' % self.xc)

    def create_xc_contributions(self, type):
        """General function for creating exchange or correlation energies"""
        assert type in ['exch', 'corr']
        err = 'bee_type %i not implemented' % self.bee_type

        if type == 'exch':
            if self.bee_type == 1:
                out = self.beefvdw_energy_contribs_x()
            elif self.bee_type in [2, 3]:
                out = self.mbeef_exchange_energy_contribs()
            else:
                raise NotImplementedError(err)
        else:
            if self.bee_type == 1:
                out = self.beefvdw_energy_contribs_c()
            elif self.bee_type == 2:
                out = np.array([])
            elif self.bee_type == 3:
                out = self.mbeefvdw_energy_contribs_c()
            else:
                raise NotImplementedError(err)
        return out

    def get_non_xc_total_energies(self):
        """Compile non-XC total energy contributions"""
        if self.e_dft is None:
            self.e_dft = self.calc.get_potential_energy()
        if self.e0 is None:
            from gpaw.xc.kernel import XCNull
            xc_null = XC(XCNull())
            self.e0 = self.e_dft + self.calc.get_xc_difference(xc_null)
        assert isinstance(self.e_dft, float)
        assert isinstance(self.e0, float)

    def mbeef_exchange_energy_contribs(self):
        """Legendre polynomial exchange contributions to mBEEF Etot"""
        self.get_non_xc_total_energies()
        e_x = np.zeros((self.max_order, self.max_order))
        for p1 in range(self.max_order):  # alpha
            for p2 in range(self.max_order):  # s2
                pars_i = np.array([1, self.trans[0], p2, 1.0])
                pars_j = np.array([1, self.trans[1], p1, 1.0])
                pars = np.hstack((pars_i, pars_j))
                x = XC('2D-MGGA', pars)
                e_x[p1, p2] = (self.e_dft +
                               self.calc.get_xc_difference(x) - self.e0)
                del x
        return e_x

    def beefvdw_energy_contribs_x(self):
        """Legendre polynomial exchange contributions to BEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_pbe = (self.e_dft + self.calc.get_xc_difference('GGA_C_PBE') -
                 self.e0)

        exch = np.zeros(30)
        for p in range(30):
            pars = [4, 0, p, 1.0]
            bee = XC('BEE2', pars)
            exch[p] = (self.e_dft + self.calc.get_xc_difference(bee) -
                       self.e0 - e_pbe)
            del bee
        return exch

    def beefvdw_energy_contribs_c(self):
        """LDA and PBE correlation contributions to BEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_lda = self.e_dft + self.calc.get_xc_difference('LDA_C_PW') - self.e0
        e_pbe = self.e_dft + self.calc.get_xc_difference('GGA_C_PBE') - self.e0
        corr = np.array([e_lda, e_pbe])
        return corr

    def mbeefvdw_energy_contribs_c(self):
        """LDA, PBEsol, and nl2 correlation contributions to mBEEF-vdW Etot"""
        self.get_non_xc_total_energies()
        e_lda = self.e_dft + self.calc.get_xc_difference('LDA_C_PW') - self.e0
        e_sol = (self.e_dft + self.calc.get_xc_difference('GGA_C_PBE_SOL') -
                 self.e0)
        vdwdf2 = VDWFunctional('vdW-DF2')
        self.calc.get_xc_difference(vdwdf2)
        e_nl2 = vdwdf2.get_Ecnl() * Hartree
        corr = np.array([e_lda, e_sol, e_nl2])
        return corr



import sys
import numpy as np

from ase.units import _hbar, _c, _e, _me, Hartree, Bohr
from gpaw import __version__ as version
from gpaw.utilities.folder import Folder


def get_folded_spectrum(
        exlist=None,
        emin=None,
        emax=None,
        de=None,
        energyunit='eV',
        folding='Gauss',
        width=0.08,  # Gauss/Lorentz width
        form='r'):
    """Return folded spectrum."""

    x = []
    y = []
    for ex in exlist:
        x.append(ex.get_energy() * Hartree)
        y.append(ex.get_oscillator_strength(form))

    if energyunit == 'nm':
        # transform to experimentally used wavelength [nm]
        x = 1.e+9 * 2 * np.pi * _hbar * _c / _e / np.array(x)
    elif energyunit != 'eV':
        raise RuntimeError('currently only eV and nm are supported')

    if folding is None:
        return x, y
    else:
        return Folder(width, folding).fold(x, y, de, emin, emax)


def spectrum(exlist=None,
             filename=None,
             emin=None,
             emax=None,
             de=None,
             energyunit='eV',
             folding='Gauss',
             width=0.08,  # Gauss/Lorentz width
             comment=None,
             form='r'):
    """Write out a folded spectrum.

    Parameters
    ----------
    exlist: ExcitationList
    filename:
        File name for the output file, STDOUT if not given
    emin:
        min. energy, set to cover all energies if not given
    emax:
        max. energy, set to cover all energies if not given
    de:
        energy spacing
    energyunit: {'eV', 'nm'}
        Energy unit, default 'eV'
    folding: {'Gauss', 'Lorentz'}
    width:
        folding width in terms of the chosen energyunit
    """

    # output
    out = sys.stdout
    if filename is not None:
        with open(filename, 'w') as out:
            if comment:
                print('#', comment, file=out)

            print('# Photoabsorption spectrum from linear response TD-DFT',
                  file=out)
            print('# GPAW version:', version, file=out)
            if folding is not None:  # fold the spectrum
                print('# %s folded, width=%g [%s]' % (folding, width,
                                                      energyunit), file=out)
            if form == 'r':
                out.write('# length form')
            else:
                assert form == 'v'
                out.write('# velocity form')
            print('# om [%s]     osz          osz x       osz y       osz z'
                  % energyunit, file=out)

            energies, values = get_folded_spectrum(exlist, emin, emax, de,
                                                   energyunit, folding,
                                                   width, form)

            for e, val in zip(energies, values):
                print('%10.5f %12.7e %12.7e %11.7e %11.7e' %
                      (e, val[0], val[1], val[2], val[3]), file=out)


def get_adsorbance_pre_factor(atoms):
    """Return the absorbance pre-factor for solids. Unit m^-1.

    Use this factor to multiply the folded oscillatior strength
    obtained from spectrum()

    Robert C. Hilborn, Am. J. Phys. 50, 982 (1982)
    """
    return np.pi * _e**2 / 2. / _me / _c


def rotatory_spectrum(exlist=None,
                      filename=None,
                      emin=None,
                      emax=None,
                      de=None,
                      energyunit='eV',
                      folding='Gauss',
                      width=0.08,  # Gauss/Lorentz width
                      comment=None):
    """Write out a folded rotatory spectrum.

    See spectrum() for explanation of the parameters.
    """

    # output
    out = sys.stdout
    if filename is not None:
        with open(filename, 'w') as out:
            if comment:
                print('#', comment, file=out)

            print('# Rotatory spectrum from linear response TD-DFT', file=out)
            print('# GPAW version:', version, file=out)
            if folding is not None:  # fold the spectrum
                print('# %s folded, width=%g [%s]' % (folding, width,
                                                      energyunit), file=out)
            print('# om [%s]     R [cgs]'
                  % energyunit, file=out)

            x = []
            y = []
            for ex in exlist:
                x.append(ex.get_energy() * Hartree)
                y.append(ex.get_rotatory_strength())

            if energyunit == 'nm':
                # transform to experimentally used wavelength [nm]
                x = 1.e+9 * 2 * np.pi * _hbar * _c / _e / np.array(x)
                y = np.array(y)
            elif energyunit != 'eV':
                raise RuntimeError('currently only eV and nm are supported')

            if folding is None:
                energies, values = x, y
            else:
                energies, values = Folder(width, folding).fold(x, y, de,
                                                               emin, emax)
            for e, val in zip(energies, values):
                print('%10.5f %12.7e' % (e, val), file=out)


class Writer(Folder):
    """Object to write data to a file"""
    def __init__(self, folding=None, width=0.08):
        """Evaluate the polarizability from sum over states.

        Parameters
        ----------
        folding : string or None(default)
          Name of the folding function, see gpaw/folder.py for options.
          None means do not fold at all.
        width : float
          The width to be transferred to the folding function.
        """
        self.folding = folding
        if folding is not None:
            Folder.__init__(self, width, folding)

    def write(self, filename=None,
              emin=None, emax=None, de=None,
              comment=None):

        out = sys.stdout
        if filename is not None:
            with open(filename, 'w') as out:
                print('#', self.title, file=out)
                print('# GPAW version:', version, file=out)
                if comment:
                    print('#', comment, file=out)
                if self.folding is not None:
                    print('# %s folded, width=%g [eV]' % (self.folding,
                          self.width), file=out)
                    energies, values = self.fold(self.energies, self.values,
                                                 de, emin, emax)
                else:
                    energies, values = self.energies, self.values

                print('#', self.fields, file=out)

                for e, val in zip(energies, values):
                    string = '%10.5f' % e
                    for vf in val:
                        string += ' %12.7e' % vf
                    print(string, file=out)


def polarizability(exlist, omega, form='v',
                   tensor=False, index=0):
    """Evaluate the polarizability from sum over states.

    Parameters
    ----------
    exlist: ExcitationList
    omega:
        Photon energy (eV)
    form: {'v', 'r'}
        Form of the dipole matrix element
    index: {0, 1, 2, 3}
        0: averaged, 1,2,3:alpha_xx, alpha_yy, alpha_zz
    tensor:
        if True returns alpha_ij, i,j=x,y,z
        index is ignored

    Returns
    -------
    alpha:
        Unit (e^2 Angstrom^2 / eV).
        Multiply with Bohr * Ha to get (Angstrom^3)
    """
    if tensor:
        alpha = np.zeros(np.array(omega).shape + (3, 3), dtype=complex)
        for ex in exlist:
            alpha += ex.get_dipole_tensor(form=form) / (
                (ex.energy * Hartree)**2 - omega**2)
    else:
        alpha = np.zeros_like(omega)
        for ex in exlist:
            alpha += ex.get_oscillator_strength(form=form)[index] / (
                (ex.energy * Hartree)**2 - omega**2)

    return alpha * Bohr**2 * Hartree



# see changeset 4891
import pytest
import numpy as np
from ase import Atoms
from gpaw import GPAW
from gpaw.mpi import world


def test_parallel_kptpar(in_tmp_dir):
    a = 2.5
    H = Atoms('H', cell=[a, a, a], pbc=True)

    energy_tolerance = 0.0006

    base_params = dict(mode='fd', kpts=[6, 6, 1], spinpol=True)
    if world.size >= 3:
        calc = GPAW(**base_params,
                    parallel={'domain': world.size},
                    txt='H-a.txt')
        H.calc = calc
        e1 = H.get_potential_energy()
        assert H.calc.wfs.kd.comm.size == 1

        assert e1 == pytest.approx(-2.23708481, abs=energy_tolerance)

        if world.rank < 3:
            comm = world.new_communicator(np.array([0, 1, 2]))
            H.calc = GPAW(**base_params,
                          communicator=comm,
                          txt='H-b.txt')
            e2 = H.get_potential_energy()
            assert H.calc.wfs.kd.comm.size == 3
            assert e2 == pytest.approx(e1, abs=5e-9)
        else:
            comm = world.new_communicator(np.array(range(3, world.size)))
            H.calc = GPAW(**base_params,
                          communicator=comm,
                          parallel={'kpt': comm.size},
                          txt='H-b2.txt')
            e2 = H.get_potential_energy()
            assert e2 == pytest.approx(e1, abs=5e-9)



import pytest
from ase import Atom, Atoms

from gpaw import GPAW


def test_aedensity():
    h = 0.21  # gridspacing
    a = [6.5, 6.5, 7.7]  # unit cell
    d = 2.3608  # experimental bond length

    NaCl = Atoms([Atom('Na', [0, 0, 0]),
                  Atom('Cl', [0, 0, d])],
                 pbc=False, cell=a)
    NaCl.center()
    calc = GPAW(mode='fd',
                h=h,
                xc='LDA',
                nbands=5,
                setups={'Na': '1'},
                convergence={'eigenstates': 1e-6},
                spinpol=1)

    NaCl.calc = calc
    e = NaCl.get_potential_energy()

    assert e == pytest.approx(-4.907, abs=0.002)

    dv = NaCl.get_volume() / calc.get_number_of_grid_points().prod()
    nt1 = calc.get_pseudo_density(gridrefinement=1)
    Zt1 = nt1.sum() * dv
    nt2 = calc.get_pseudo_density(gridrefinement=2)
    Zt2 = nt2.sum() * dv / 8
    print('Integral of pseudo density:', Zt1, Zt2)
    assert Zt1 == pytest.approx(Zt2, abs=1e-12)

    for gridrefinement in [1, 2, 4]:
        n = calc.get_all_electron_density(gridrefinement=gridrefinement)
        Z = n.sum() * dv / gridrefinement**3
        print('Integral of all-electron density:', Z)
        assert Z == pytest.approx(28, abs=1e-5)



from __future__ import annotations

import warnings
from pathlib import Path
from types import SimpleNamespace
from typing import IO, Any, Union

import numpy as np
from ase import Atoms
from ase.units import Bohr, Ha
from gpaw import __version__
from gpaw.core import UGArray
from gpaw.dos import DOSCalculator
from gpaw.mpi import world, synchronize_atoms
from gpaw.new import Timer, cached_property
from gpaw.new.builder import builder as create_builder
from gpaw.new.calculation import (DFTCalculation, DFTState,
                                  ReuseWaveFunctionsError, units)
from gpaw.new.gpw import read_gpw, write_gpw
from gpaw.new.input_parameters import (DeprecatedParameterWarning,
                                       InputParameters)
from gpaw.new.logger import Logger
from gpaw.new.pw.fulldiag import diagonalize
from gpaw.new.xc import create_functional
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import pack
from gpaw.utilities.memory import maxrss


def GPAW(filename: Union[str, Path, IO[str]] = None,
         txt: str | Path | IO[str] | None = '?',
         communicator=None,
         **kwargs) -> ASECalculator:
    """Create ASE-compatible GPAW calculator."""
    if txt == '?':
        txt = '-' if filename is None else None

    parallel = kwargs.get('parallel', {})
    comm = parallel.pop('world', None)
    if comm is None:
        comm = communicator or world
    else:
        warnings.warn(('Please use communicator=... '
                       'instead of parallel={''world'': ...}'),
                      DeprecatedParameterWarning)
    log = Logger(txt, comm)

    if filename is not None:
        if not {'parallel'}.issuperset(kwargs):
            illegal = set(kwargs) - {'parallel'}
            raise ValueError('Illegal arguments when reading from a file: '
                             f'{illegal}')
        atoms, calculation, params, _ = read_gpw(filename,
                                                 log=log,
                                                 parallel=parallel)
        return ASECalculator(params,
                             log=log, calculation=calculation, atoms=atoms)

    params = InputParameters(kwargs)
    write_header(log, params)
    return ASECalculator(params, log=log)


def write_header(log, params):
    from gpaw.io.logger import write_header as header
    log(f'#  __  _  _\n# | _ |_)|_||  |\n# |__||  | ||/\\| - {__version__}\n')
    header(log, log.comm)
    log('---')
    with log.indent('input parameters:'):
        log(**dict(params.items()))


def compare_atoms(a1: Atoms, a2: Atoms) -> set[str]:
    if a1 is a2:
        return set()

    if len(a1.numbers) != len(a2.numbers) or (a1.numbers != a2.numbers).any():
        return {'numbers'}

    if (a1.pbc != a2.pbc).any():
        return {'pbc'}

    if abs(a1.cell - a2.cell).max() > 0.0:
        return {'cell'}

    if abs(a1.positions - a2.positions).max() > 0.0:
        return {'positions'}

    return set()


class ASECalculator:
    """This is the ASE-calculator frontend for doing a GPAW calculation."""

    name = 'gpaw'

    def __init__(self,
                 params: InputParameters,
                 *,
                 log: Logger,
                 calculation=None,
                 atoms=None):
        self.params = params
        self.log = log
        self.comm = log.comm
        self.calculation = calculation

        self.atoms = atoms
        self.timer = Timer()

    def __repr__(self):
        params = []
        for key, value in self.params.items():
            val = repr(value)
            if len(val) > 40:
                val = '...'
            params.append((key, val))
        p = ', '.join(f'{key}: {val}' for key, val in params)
        return f'ASECalculator({p})'

    def calculate_property(self,
                           atoms: Atoms | None,
                           prop: str) -> Any:
        """Calculate (if not already calculated) a property.

        The ``prop`` string must be one of

        * energy
        * forces
        * stress
        * magmom
        * magmoms
        * dipole
        """
        if atoms is None:
            atoms = self.atoms
        else:
            synchronize_atoms(atoms, self.comm)
        assert atoms is not None

        if self.calculation is not None:
            changes = compare_atoms(self.atoms, atoms)
            if changes & {'numbers', 'pbc', 'cell'}:
                if 'numbers' not in changes:
                    # Remember magmoms if there are any:
                    magmom_a = self.calculation.results.get('magmoms')
                    if magmom_a is not None and magmom_a.any():
                        atoms = atoms.copy()
                        assert atoms is not None  # MYPY: why is this needed?
                        atoms.set_initial_magnetic_moments(magmom_a)

                if changes & {'numbers', 'pbc'}:
                    # Start from scratch:
                    self.calculation = None
                else:
                    ibzwfs = self.calculation.state.ibzwfs
                    kpt_parallel_only = (ibzwfs.band_comm.size == 1 and
                                         ibzwfs.domain_comm.size == 1)
                    if kpt_parallel_only:
                        try:
                            self.create_new_calculation_from_old(atoms)
                        except ReuseWaveFunctionsError:
                            self.calculation = None
                        else:
                            self.converge()
                            changes = set()
                    else:
                        # Not implemented: just start from scratch
                        self.calculation = None

        if self.calculation is None:
            self.create_new_calculation(atoms)
            assert self.calculation is not None
            self.converge()
        elif changes:
            self.move_atoms(atoms)
            self.converge()

        if prop not in self.calculation.results:
            if prop == 'forces':
                with self.timer('Forces'):
                    self.calculation.forces()
            elif prop == 'stress':
                with self.timer('Stress'):
                    self.calculation.stress()
            elif prop == 'dipole':
                self.calculation.dipole()
            else:
                raise KeyError('Unknown property:', prop)

        return self.calculation.results[prop] * units[prop]

    def get_property(self,
                     name: str,
                     atoms: Atoms | None = None,
                     allow_calculation: bool = True) -> Any:
        if not allow_calculation and name not in self.calculation.results:
            return None
        if atoms is None:
            atoms = self.atoms
        return self.calculate_property(atoms, name)

    @property
    def results(self):
        if self.calculation is None:
            return {}
        return {name: value * units[name]
                for name, value in self.calculation.results.items()}

    def create_new_calculation(self, atoms: Atoms) -> None:
        with self.timer('Init'):
            self.calculation = DFTCalculation.from_parameters(
                atoms, self.params, self.comm, self.log)
        self.atoms = atoms.copy()

    def create_new_calculation_from_old(self, atoms: Atoms) -> None:
        with self.timer('Morph'):
            self.calculation = self.calculation.new(
                atoms, self.params, self.log)
        self.atoms = atoms.copy()

    def move_atoms(self, atoms):
        with self.timer('Move'):
            self.calculation = self.calculation.move_atoms(atoms)
        self.atoms = atoms.copy()

    def converge(self):
        """Iterate to self-consistent solution.

        Will also calculate "cheap" properties: energy, magnetic moments
        and dipole moment.
        """
        with self.timer('SCF'):
            self.calculation.converge(calculate_forces=self._calculate_forces)

        # Calculate all the cheap things:
        self.calculation.energies()
        self.calculation.dipole()
        self.calculation.magmoms()

        self.calculation.write_converged()

    def _calculate_forces(self) -> Array2D:  # units: Ha/Bohr
        """Helper method for force-convergence criterium."""
        with self.timer('Forces'):
            self.calculation.forces(silent=True)
        return self.calculation.results['forces']

    def __del__(self):
        self.log('---')
        self.timer.write(self.log)
        try:
            mib = maxrss() / 1024**2
            self.log(f'\nMax RSS: {mib:.3f}  # MiB')
        except NameError:
            pass

    def get_potential_energy(self,
                             atoms: Atoms | None = None,
                             force_consistent: bool = False) -> float:
        return self.calculate_property(atoms,
                                       'free_energy' if force_consistent else
                                       'energy')

    def get_forces(self, atoms: Atoms | None = None) -> Array2D:
        return self.calculate_property(atoms, 'forces')

    def get_stress(self, atoms: Atoms | None = None) -> Array1D:
        return self.calculate_property(atoms, 'stress')

    def get_dipole_moment(self, atoms: Atoms | None = None) -> Array1D:
        return self.calculate_property(atoms, 'dipole')

    def get_magnetic_moment(self, atoms: Atoms | None = None) -> float:
        return self.calculate_property(atoms, 'magmom')

    def get_magnetic_moments(self, atoms: Atoms | None = None) -> Array1D:
        return self.calculate_property(atoms, 'magmoms')

    def write(self, filename, mode=''):
        """Write calculator object to a file.

        Parameters
        ----------
        filename:
            File to be written
        mode:
            Write mode. Use ``mode='all'``
            to include wave functions in the file.
        """
        self.log(f'# Writing to {filename} (mode={mode!r})\n')

        write_gpw(filename, self.atoms, self.params,
                  self.calculation, skip_wfs=mode != 'all')

    # Old API:

    implemented_properties = ['energy', 'free_energy',
                              'forces', 'stress',
                              'dipole', 'magmom', 'magmoms']

    def new(self, **kwargs) -> ASECalculator:
        kwargs = {**dict(self.params.items()), **kwargs}
        return GPAW(**kwargs)

    def get_pseudo_wave_function(self, band, kpt=0, spin=0,
                                 periodic=False) -> Array3D:
        state = self.calculation.state
        wfs = state.ibzwfs.get_wfs(spin=spin, kpt=kpt, n1=band, n2=band + 1)
        basis = getattr(self.calculation.scf_loop.hamiltonian, 'basis', None)
        grid = state.density.nt_sR.desc
        wfs = wfs.to_uniform_grid_wave_functions(grid, basis)
        psit_R = wfs.psit_nX[0]
        if not psit_R.desc.pbc.all():
            psit_R = psit_R.to_pbc_grid()
        if periodic:
            psit_R.multiply_by_eikr(-psit_R.desc.kpt_c)
        return psit_R.data * Bohr**-1.5

    def get_atoms(self):
        atoms = self.atoms.copy()
        atoms.calc = self
        return atoms

    def get_fermi_level(self) -> float:
        state = self.calculation.state
        fl = state.ibzwfs.fermi_levels * Ha
        assert len(fl) == 1
        return fl[0]

    def get_fermi_levels(self) -> float:
        state = self.calculation.state
        fl = state.ibzwfs.fermi_levels * Ha
        assert len(fl) == 2
        return fl

    def get_homo_lumo(self, spin: int = None) -> Array1D:
        state = self.calculation.state
        return state.ibzwfs.get_homo_lumo(spin) * Ha

    def get_number_of_electrons(self):
        state = self.calculation.state
        return state.ibzwfs.nelectrons

    def get_number_of_bands(self):
        state = self.calculation.state
        return state.ibzwfs.nbands

    def get_number_of_grid_points(self):
        return self.calculation.state.density.nt_sR.desc.size

    def get_effective_potential(self, spin=0):
        assert spin == 0
        vt_R = self.calculation.state.potential.vt_sR[spin]
        return vt_R.to_pbc_grid().gather(broadcast=True).data * Ha

    def get_electrostatic_potential(self):
        density = self.calculation.state.density
        potential, _ = self.calculation.pot_calc.calculate(density)
        vHt_x = potential.vHt_x
        if isinstance(vHt_x, UGArray):
            return vHt_x.gather(broadcast=True).to_pbc_grid().data * Ha

        return vHt_x.ifft(
            grid=self.calculation.pot_calc.fine_grid).data * Ha

    def get_atomic_electrostatic_potentials(self):
        return self.calculation.electrostatic_potential().atomic_potentials()

    def get_electrostatic_corrections(self):
        return self.calculation.electrostatic_potential().atomic_corrections()

    def get_pseudo_density(self,
                           spin=None,
                           gridrefinement=1,
                           broadcast=True) -> Array3D:
        assert spin is None
        nt_sr = self.calculation.densities().pseudo_densities(
            grid_refinement=gridrefinement)
        return nt_sr.gather(broadcast=broadcast).data.sum(0)

    def get_all_electron_density(self,
                                 spin=None,
                                 gridrefinement=1,
                                 broadcast=True,
                                 skip_core=False):
        assert spin is None
        n_sr = self.calculation.densities().all_electron_densities(
            grid_refinement=gridrefinement,
            skip_core=skip_core)
        return n_sr.gather(broadcast=broadcast).data.sum(0)

    def get_eigenvalues(self, kpt=0, spin=0, broadcast=True):
        state = self.calculation.state
        eig_n = state.ibzwfs.get_eigs_and_occs(k=kpt, s=spin)[0] * Ha
        if broadcast:
            if self.comm.rank != 0:
                eig_n = np.empty(state.ibzwfs.nbands)
            self.comm.broadcast(eig_n, 0)
        return eig_n

    def get_occupation_numbers(self, kpt=0, spin=0, broadcast=True):
        state = self.calculation.state
        weight = state.ibzwfs.ibz.weight_k[kpt] * state.ibzwfs.spin_degeneracy
        occ_n = state.ibzwfs.get_eigs_and_occs(k=kpt, s=spin)[1] * weight
        if broadcast:
            if self.comm.rank != 0:
                occ_n = np.empty(state.ibzwfs.nbands)
            self.comm.broadcast(occ_n, 0)
        return occ_n

    def get_reference_energy(self):
        return self.calculation.setups.Eref * Ha

    def get_number_of_iterations(self):
        return self.calculation.scf_loop.niter

    def get_bz_k_points(self):
        state = self.calculation.state
        return state.ibzwfs.ibz.bz.kpt_Kc.copy()

    def get_ibz_k_points(self):
        state = self.calculation.state
        return state.ibzwfs.ibz.kpt_kc.copy()

    def get_orbital_magnetic_moments(self):
        """Return the orbital magnetic moment vector for each atom."""
        from gpaw.new.orbmag import get_orbmag_from_calc
        return get_orbmag_from_calc(self)

    def calculate(self, atoms, properties=None, system_changes=None):
        if properties is None:
            properties = ['energy']

        for name in properties:
            self.calculate_property(atoms, name)
        # self.get_potential_energy(atoms)

    @cached_property
    def wfs(self):
        from gpaw.new.backwards_compatibility import FakeWFS
        return FakeWFS(self.calculation, self.atoms)

    @property
    def density(self):
        from gpaw.new.backwards_compatibility import FakeDensity
        return FakeDensity(self.calculation)

    @property
    def hamiltonian(self):
        from gpaw.new.backwards_compatibility import FakeHamiltonian
        return FakeHamiltonian(self.calculation)

    @property
    def spos_ac(self):
        return self.atoms.get_scaled_positions()

    @property
    def world(self):
        return self.comm

    @property
    def setups(self):
        return self.calculation.setups

    @property
    def initialized(self):
        return self.calculation is not None

    def get_xc_functional(self):
        return self.calculation.pot_calc.xc.name

    def get_xc_difference(self, xcparams):
        """Calculate non-selfconsistent XC-energy difference."""
        dft = self.calculation
        pot_calc = dft.pot_calc
        state = dft.state
        density = dft.state.density
        xc = create_functional(xcparams, pot_calc.fine_grid)
        if xc.type == 'MGGA' and density.taut_sR is None:
            state.ibzwfs.make_sure_wfs_are_read_from_gpw_file()
            if isinstance(state.ibzwfs.wfs_qs[0][0].psit_nX, SimpleNamespace):
                params = InputParameters(dict(self.params.items()))
                builder = create_builder(self.atoms, params, self.comm)
                basis_set = builder.create_basis_set()
                ibzwfs = builder.create_ibz_wave_functions(
                    basis_set, state.potential, log=dft.log)
                ibzwfs.fermi_levels = state.ibzwfs.fermi_levels
                state.ibzwfs = ibzwfs
                dft.scf_loop.update_density_and_potential = False
                dft.converge()
            density.update_ked(state.ibzwfs)
        exct = pot_calc.calculate_non_selfconsistent_exc(
            xc, density.nt_sR, density.taut_sR)
        dexc = 0.0
        for a, D_sii in state.density.D_asii.items():
            setup = self.setups[a]
            dexc += xc.calculate_paw_correction(
                setup,
                np.array([pack(D_ii) for D_ii in D_sii.real]))
        return (exct + dexc - state.potential.energies['xc']) * Ha

    def diagonalize_full_hamiltonian(self,
                                     nbands: int = None,
                                     scalapack=None,
                                     expert: bool = None) -> None:
        if expert is not None:
            warnings.warn('Ignoring deprecated "expert" argument',
                          DeprecationWarning)
        state = self.calculation.state
        ibzwfs = diagonalize(state.potential,
                             state.ibzwfs,
                             self.calculation.scf_loop.occ_calc,
                             nbands,
                             self.calculation.pot_calc.xc)
        self.calculation.state = DFTState(ibzwfs,
                                          state.density,
                                          state.potential)
        nbands = ibzwfs.nbands
        self.params.nbands = nbands
        self.params.keys.append('nbands')

    def gs_adapter(self):
        from gpaw.response.groundstate import ResponseGroundStateAdapter
        return ResponseGroundStateAdapter(self)

    def fixed_density(self, txt='-', **kwargs):
        kwargs = {**dict(self.params.items()), **kwargs}
        params = InputParameters(kwargs)
        log = Logger(txt, self.comm)
        builder = create_builder(self.atoms, params, self.comm)
        basis_set = builder.create_basis_set()
        state = self.calculation.state
        comm1 = state.ibzwfs.kpt_band_comm
        comm2 = builder.communicators['D']
        potential = state.potential.redist(
            builder.grid,
            builder.electrostatic_potential_desc,
            builder.atomdist,
            comm1, comm2)
        density = state.density.redist(builder.grid,
                                       builder.interpolation_desc,
                                       builder.atomdist,
                                       comm1, comm2)
        ibzwfs = builder.create_ibz_wave_functions(basis_set, potential,
                                                   log=log)
        ibzwfs.fermi_levels = state.ibzwfs.fermi_levels
        state = DFTState(ibzwfs, density, potential)
        scf_loop = builder.create_scf_loop()
        scf_loop.update_density_and_potential = False

        calculation = DFTCalculation(
            state,
            builder.setups,
            scf_loop,
            SimpleNamespace(fracpos_ac=self.calculation.fracpos_ac,
                            poisson_solver=None),
            log)

        calculation.converge()

        return ASECalculator(params,
                             log=log,
                             calculation=calculation,
                             atoms=self.atoms)

    def initialize(self, atoms):
        self.create_new_calculation(atoms)

    def converge_wave_functions(self):
        self.calculation.state.ibzwfs.make_sure_wfs_are_read_from_gpw_file()

    def get_number_of_spins(self):
        return self.calculation.state.density.ndensities

    @property
    def parameters(self):
        return self.params

    def dos(self,
            soc: bool = False,
            theta: float = 0.0,  # degrees
            phi: float = 0.0,  # degrees
            shift_fermi_level: bool = True) -> DOSCalculator:
        """Create DOS-calculator.

        Default is to ``shift_fermi_level`` to 0.0 eV.  For ``soc=True``,
        angles can be given in degrees.
        """
        return DOSCalculator.from_calculator(
            self, soc=soc,
            theta=theta, phi=phi,
            shift_fermi_level=shift_fermi_level)

    def band_structure(self):
        """Create band-structure object for plotting."""
        from ase.spectrum.band_structure import get_band_structure
        return get_band_structure(calc=self)

    @property
    def symmetry(self):
        return self.calculation.state.ibzwfs.ibz.symmetries.symmetry



"""Kohn-Sham single particle excitations realated objects.

"""
import sys
import json
import numpy as np
from copy import copy

from ase.units import Bohr, Hartree, alpha

import gpaw.mpi as mpi
from gpaw.utilities import packed_index
from gpaw.lrtddft.excitation import Excitation, ExcitationList, get_filehandle
from gpaw.pair_density import PairDensity
from gpaw.fd_operators import Gradient
from gpaw.utilities.tools import coordinates


class KSSingles(ExcitationList):
    """Kohn-Sham single particle excitations

    Input parameters:

    calculator:
      the calculator object after a ground state calculation

    nspins:
      number of spins considered in the calculation
      Note: Valid only for unpolarised ground state calculation

    eps:
      Minimal occupation difference for a transition (default 0.001)

    istart:
      First occupied state to consider
    jend:
      Last unoccupied state to consider
    energy_range:
      The energy range [emin, emax] or emax for KS transitions to use as basis
    """
    def __init__(self,
                 restrict={},
                 log=None,
                 txt=None):
        ExcitationList.__init__(self, log=log, txt=txt)
        self.world = mpi.world

        self.restrict = KSSRestrictor()
        self.restrict.update(restrict)

    def calculate(self, atoms, nspins=None):
        calculator = atoms.calc
        self.calculator = calculator

        # LCAO calculation requires special actions
        self.lcao = calculator.wfs.mode == 'lcao'

        # deny hybrids as their empty states are wrong
#        gsxc = calculator.hamiltonian.xc
#        hybrid = hasattr(gsxc, 'hybrid') and gsxc.hybrid > 0.0
#        assert(not hybrid)

        # ensure correctly initialized wave functions
        calculator.converge_wave_functions()
        self.world = calculator.wfs.world

        # parallelization over bands not yet supported
        assert calculator.wfs.bd.comm.size == 1

        # do the evaluation
        self.select(nspins)

        trkm = self.get_trk()
        self.log('KSS {0} transitions (restrict={1})'.format(
            len(self), self.restrict))
        self.log('KSS TRK sum %g (%g,%g,%g)' %
                 (np.sum(trkm) / 3., trkm[0], trkm[1], trkm[2]))
        pol = self.get_polarizabilities(lmax=3)
        self.log('KSS polarisabilities(l=0-3) %g, %g, %g, %g' %
                 tuple(pol.tolist()))
        return self

    @staticmethod
    def emin_emax(energy_range):
        emin = -sys.float_info.max
        emax = sys.float_info.max
        if energy_range is not None:
            try:
                emin, emax = energy_range
                emin /= Hartree
                emax /= Hartree
            except TypeError:
                emax = energy_range / Hartree
        return emin, emax

    def select(self, nspins=None):
        """Select KSSingles according to the given criterium."""

        # criteria
        emin, emax = self.restrict.emin_emax()
        istart = self.restrict['istart']
        jend = self.restrict['jend']
        eps = self.restrict['eps']

        if not hasattr(self, 'calculator'):  # I'm read from a file
            # throw away all not needed entries
            for i, ks in reversed(list(enumerate(self))):
                if ((ks.fij / ks.weight) <= eps or
                   ks.i < istart or ks.j > jend or
                   ks.energy < emin or ks.energy > emax):
                    del self[i]
            return None

        paw = self.calculator
        wfs = paw.wfs
        self.dtype = wfs.dtype
        self.kpt_u = wfs.kpt_u

        if not self.lcao and self.kpt_u[0].psit_nG is None:
            raise RuntimeError('No wave functions in calculator!')

        # here, we need to take care of the spins also for
        # closed shell systems (Sz=0)
        # vspin is the virtual spin of the wave functions,
        #       i.e. the spin used in the ground state calculation
        # pspin is the physical spin of the wave functions
        #       i.e. the spin of the excited states
        self.nvspins = wfs.nspins
        self.npspins = wfs.nspins
        fijscale = 1
        ispins = [0]
        nks = wfs.kd.nibzkpts * wfs.kd.nspins
        if self.nvspins < 2:
            if (nspins or 0) > self.nvspins:
                self.npspins = nspins
                fijscale = 0.5
                ispins = [0, 1]
                nks *= 2

        kpt_comm = self.calculator.wfs.kd.comm
        nbands = len(self.kpt_u[0].f_n)

        # select
        take = np.zeros((nks, nbands, nbands), dtype=int)
        u = 0
        for ispin in ispins:
            for k in range(wfs.kd.nibzkpts):
                q = k - wfs.kd.k0
                for s in range(wfs.nspins):
                    if q >= 0 and q < wfs.kd.mynk:
                        kpt = wfs.kpt_qs[q][s]
                        for i in range(nbands):
                            for j in range(i + 1, nbands):
                                fij = (kpt.f_n[i] - kpt.f_n[j]) / kpt.weight
                                epsij = kpt.eps_n[j] - kpt.eps_n[i]
                                if (fij > eps and
                                    epsij >= emin and epsij < emax and
                                        i >= istart and j <= jend):
                                    take[u, i, j] = 1
                    u += 1
        kpt_comm.sum(take)

        self.log()
        self.log('Kohn-Sham single transitions')
        self.log()

        # calculate in parallel
        u = 0
        for ispin in ispins:
            for k in range(wfs.kd.nibzkpts):
                q = k - wfs.kd.k0
                for s in range(wfs.kd.nspins):
                    for i in range(nbands):
                        for j in range(i + 1, nbands):
                            if take[u, i, j]:
                                if q >= 0 and q < wfs.kd.mynk:
                                    kpt = wfs.kpt_qs[q][s]
                                    pspin = max(kpt.s, ispin)
                                    self.append(
                                        KSSingle(i, j, pspin, kpt, paw,
                                                 fijscale=fijscale,
                                                 dtype=self.dtype))
                                else:
                                    self.append(KSSingle(i, j, pspin=0,
                                                         kpt=None, paw=paw,
                                                         dtype=self.dtype))
                    u += 1

        # distribute
        for kss in self:
            kss.distribute()

    @classmethod
    def read(cls, filename=None, fh=None, restrict={}, log=None):
        """Read myself from a file"""
        assert (filename is not None) or (fh is not None)

        def fail(f):
            raise RuntimeError(f.name + ' does not contain ' +
                               cls.__class__.__name__ + ' data')
        if fh is None:
            f = get_filehandle(cls, filename)

            # there can be other information, i.e. the LrTDDFT header
            try:
                content = f.read()
                f.seek(content.index('# KSSingles'))
                del content
                f.readline()
            except ValueError:
                fail(f)
        else:
            f = fh
            # we assume to be at the right place and read the header
            if not f.readline().strip() == '# KSSingles':
                fail(f)

        words = f.readline().split()
        n = int(words[0])
        kssl = cls(log=log)
        if len(words) == 1:
            # very old output style for real wave functions (finite systems)
            kssl.dtype = float
            restrict_from_file = {}
        else:
            if words[1].startswith('complex'):
                kssl.dtype = complex
            else:
                kssl.dtype = float
            restrict_from_file = json.loads(f.readline())
            if not isinstance(restrict_from_file, dict):  # old output style
                restrict_from_file = {'eps': restrict_from_file}
        kssl.npspins = 1
        for i in range(n):
            kss = KSSingle(string=f.readline(), dtype=kssl.dtype)
            kssl.append(kss)
            kssl.npspins = max(kssl.npspins, kss.pspin + 1)

        if fh is None:
            f.close()

        kssl.update()
        kssl.restrict.update(restrict_from_file)
        if len(restrict):
            kssl.restrict.update(restrict)
            kssl.select()

        return kssl

    def update(self):
        istart = self[0].i
        jend = 0
        npspins = 1
        nvspins = 1
        for kss in self:
            istart = min(kss.i, istart)
            jend = max(kss.j, jend)
            if kss.pspin == 1:
                npspins = 2
            if kss.spin == 1:
                nvspins = 2
        self.restrict.update({'istart': istart, 'jend': jend})
        self.npspins = npspins
        self.nvspins = nvspins

        if hasattr(self, 'energies'):
            del self.energies

    def set_arrays(self):
        if hasattr(self, 'energies'):
            return
        energies = []
        fij = []
        me = []
        mur = []
        muv = []
        magn = []
        for k in self:
            energies.append(k.energy)
            fij.append(k.fij)
            me.append(k.me)
            mur.append(k.mur)
            if k.muv is not None:
                muv.append(k.muv)
            if k.magn is not None:
                magn.append(k.magn)
        self.energies = np.array(energies)
        self.fij = np.array(fij)
        self.me = np.array(me)
        self.mur = np.array(mur)
        if len(muv):
            self.muv = np.array(muv)
        else:
            self.muv = None
        if len(magn):
            self.magn = np.array(magn)
        else:
            self.magn = None

    def write(self, filename=None, fh=None):
        """Write current state to a file.

        'filename' is the filename. If the filename ends in .gz,
        the file is automatically saved in compressed gzip format.

        'fh' is a filehandle. This can be used to write into already
        opened files.
        """
        if self.world.rank != 0:
            return

        if fh is None:
            f = get_filehandle(self, filename, mode='w')
        else:
            f = fh

        f.write('# KSSingles\n')
        f.write('{0} {1}\n'.format(len(self), np.dtype(self.dtype)))
        f.write(json.dumps(self.restrict.values) + '\n')
        for kss in self:
            f.write(kss.outstring())
        if fh is None:
            f.close()

    def overlap(self, ov_nn, other):
        """Matrix element overlaps determined from wave function overlaps.

        Parameters
        ----------
        ov_nn: array
            Wave function overlap factors from a displaced calculator.
            Index 0 corresponds to our own wavefunctions conjugated and
            index 1 to the others' wavefunctions

        Returns
        -------
        ov_pp: array
            Overlap corresponding to matrix elements.
            Index 0 corresponds to our own matrix elements conjugated and
            index 1 to the others' matrix elements
        """
        n0 = len(self)
        n1 = len(other)
        ov_pp = np.zeros((n0, n1), dtype=ov_nn.dtype)
        i1_p = [ex.i for ex in other]
        j1_p = [ex.j for ex in other]
        for p0, ex0 in enumerate(self):
            ov_pp[p0, :] = ov_nn[ex0.i, i1_p].conj() * ov_nn[ex0.j, j1_p]
        return ov_pp


class KSSRestrictor:
    """Object to handle KSSingles restrictions"""
    defaults = {'eps': 0.01,
                'istart': 0,
                'jend': sys.maxsize,
                'energy_range': None}

    def __init__(self, dictionary={}):
        self._vals = {}
        self.update(dictionary)

    def __getitem__(self, index):
        assert index in self.defaults
        return self._vals.get(index, self.defaults[index])

    def __setitem__(self, index, value):
        assert index in self.defaults
        self._vals[index] = value

    def update(self, dictionary):
        for key, value in dictionary.items():
            self[key] = value

    def emin_emax(self):
        emin = -sys.float_info.max
        emax = sys.float_info.max
        if self['energy_range'] is not None:
            try:
                emin, emax = self['energy_range']
                emin /= Hartree
                emax /= Hartree
            except TypeError:
                emax = self['energy_range'] / Hartree
        return emin, emax

    def __ge__(self, other):
        """am I less or equal restricting than the other?"""
        emin, emax = self.emin_emax()
        omin, omax = other.emin_emax()
        res = (emin <= omin) & (emax >= omax)

        res &= self['istart'] <= other['istart']
        res &= self['jend'] >= other['jend']
        res &= self['eps'] < other['eps']

        return res

    @property
    def values(self):
        return dict(self._vals)

    def __str__(self):
        return str(self.values)


class KSSingle(Excitation, PairDensity):
    """Single Kohn-Sham transition containing all its indices

      pspin=physical spin
      spin=virtual  spin, i.e. spin in the ground state calc.
      kpt=the Kpoint object
      fijscale=weight for the occupation difference::
      me  = sqrt(fij*epsij) * <i|r|j>
      mur = - <i|r|a>
      muv = - <i|nabla|a>/omega_ia with omega_ia>0
      magn = <i|[r x nabla]|a> / (2 m_e c)
    """

    def __init__(self, iidx=None, jidx=None, pspin=None, kpt=None,
                 paw=None, string=None, fijscale=1, dtype=float):
        """
        iidx: index of occupied state
        jidx: index of empty state
        pspin: physical spin
        kpt: kpoint object,
        paw: calculator,
        string: string to be initialized from
        fijscale:
        dtype: dtype of matrix elements
        """
        if string is not None:
            self.fromstring(string, dtype)
            return None

        # normal entry

        PairDensity.__init__(self, paw)
        PairDensity.initialize(self, kpt, iidx, jidx)

        self.pspin = pspin

        self.energy = 0.0
        self.fij = 0.0

        self.me = np.zeros((3), dtype=dtype)
        self.mur = np.zeros((3), dtype=dtype)
        self.muv = np.zeros((3), dtype=dtype)
        self.magn = np.zeros((3), dtype=dtype)

        self.kpt_comm = paw.wfs.kd.comm

        # leave empty if not my kpt
        if kpt is None:
            return

        wfs = paw.wfs
        gd = wfs.gd

        self.energy = kpt.eps_n[jidx] - kpt.eps_n[iidx]
        self.fij = (kpt.f_n[iidx] - kpt.f_n[jidx]) * fijscale

        # calculate matrix elements -----------

        # length form ..........................

        # course grid contribution
        # <i|r|j> is the negative of the dipole moment (because of negative
        # e- charge)
        me = - gd.calculate_dipole_moment(self.get())

        # augmentation contributions
        ma = np.zeros(me.shape, dtype=dtype)
        pos_av = paw.atoms.get_positions() / Bohr
        for a, P_ni in kpt.P_ani.items():
            Ra = pos_av[a]
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            Delta_pL = wfs.setups[a].Delta_pL
            ni = len(Pi_i)
            ma0 = 0
            ma1 = np.zeros(me.shape, dtype=me.dtype)
            for i in range(ni):
                for j in range(ni):
                    pij = Pi_i[i] * Pj_i[j]
                    ij = packed_index(i, j, ni)
                    # L=0 term
                    ma0 += Delta_pL[ij, 0] * pij
                    # L=1 terms
                    if wfs.setups[a].lmax >= 1:
                        # see spherical_harmonics.py for
                        # L=1:y L=2:z; L=3:x
                        ma1 += np.array([Delta_pL[ij, 3], Delta_pL[ij, 1],
                                         Delta_pL[ij, 2]]) * pij
            ma += np.sqrt(4 * np.pi / 3) * ma1 + Ra * np.sqrt(4 * np.pi) * ma0
        gd.comm.sum(ma)

        self.me = np.sqrt(self.energy * self.fij) * (me + ma)
        self.mur = - (me + ma)

        # velocity form .............................

        if self.lcao:
            self.wfi = _get_and_distribute_wf(wfs, iidx, kpt.k, pspin)
            self.wfj = _get_and_distribute_wf(wfs, jidx, kpt.k, pspin)

        me = np.zeros(self.mur.shape, dtype=dtype)

        # get derivatives
        dtype = self.wfj.dtype
        dwfj_cg = gd.empty((3), dtype=dtype)
        if not hasattr(gd, 'ddr'):
            gd.ddr = [Gradient(gd, c, dtype=dtype, n=2).apply
                      for c in range(3)]
        for c in range(3):
            gd.ddr[c](self.wfj, dwfj_cg[c], kpt.phase_cd)
            me[c] = gd.integrate(self.wfi.conj() * dwfj_cg[c])

        # XXX is this the best choice, maybe center of mass?
        origin = 0.5 * np.diag(paw.wfs.gd.cell_cv)

        # augmentation contributions

        # <psi_i|grad|psi_j>
        ma = np.zeros(me.shape, dtype=me.dtype)
        # Ra x <psi_i|grad|psi_j> for magnetic transition dipole
        mRa = np.zeros(me.shape, dtype=me.dtype)
        for a, P_ni in kpt.P_ani.items():
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            nabla_iiv = paw.wfs.setups[a].nabla_iiv
            ma_c = np.zeros(me.shape, dtype=me.dtype)
            for c in range(3):
                for i1, Pi in enumerate(Pi_i):
                    for i2, Pj in enumerate(Pj_i):
                        ma_c[c] += Pi * Pj * nabla_iiv[i1, i2, c]
            mRa += np.cross(paw.atoms[a].position / Bohr - origin, ma_c)
            ma += ma_c
        gd.comm.sum(ma)
        gd.comm.sum(mRa)

        self.muv = - (me + ma) / self.energy

        # magnetic transition dipole ................

        # m_ij = -(1/2c) <i|L|j> = i/2c <i|r x p|j>
        # see Autschbach et al., J. Chem. Phys., 116, 6930 (2002)

        r_cg, r2_g = coordinates(gd, origin=origin)
        magn = np.zeros(me.shape, dtype=dtype)

        # <psi_i|r x grad|psi_j>
        wfi_g = self.wfi.conj()
        for ci in range(3):
            cj = (ci + 1) % 3
            ck = (ci + 2) % 3
            magn[ci] = gd.integrate(wfi_g * r_cg[cj] * dwfj_cg[ck] -
                                    wfi_g * r_cg[ck] * dwfj_cg[cj])

        # augmentation contributions
        # <psi_i| r x nabla |psi_j>
        # = <psi_i| (r - Ra + Ra) x nabla |psi_j>
        # = <psi_i| (r - Ra) x nabla |psi_j> + Ra x <psi_i| nabla |psi_j>

        ma = np.zeros(magn.shape, dtype=magn.dtype)
        for a, P_ni in kpt.P_ani.items():
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            rxnabla_iiv = paw.wfs.setups[a].rxnabla_iiv
            for c in range(3):
                for i1, Pi in enumerate(Pi_i):
                    for i2, Pj in enumerate(Pj_i):
                        ma[c] += Pi * Pj * rxnabla_iiv[i1, i2, c]
        gd.comm.sum(ma)

        self.magn = alpha / 2. * (magn + ma + mRa)

    def distribute(self):
        """Distribute results to all cores."""
        self.spin = self.kpt_comm.sum(self.spin)
        self.pspin = self.kpt_comm.sum(self.pspin)
        self.k = self.kpt_comm.sum(self.k)
        self.weight = self.kpt_comm.sum(self.weight)
        self.energy = self.kpt_comm.sum(self.energy)
        self.fij = self.kpt_comm.sum(self.fij)

        self.kpt_comm.sum(self.me)
        self.kpt_comm.sum(self.mur)
        self.kpt_comm.sum(self.muv)
        self.kpt_comm.sum(self.magn)

    def __add__(self, other):
        """Add two KSSingles"""
        result = copy(self)
        result.me = self.me + other.me
        result.mur = self.mur + other.mur
        result.muv = self.muv + other.muv
        result.magn = self.magn + other.magn
        return result

    def __sub__(self, other):
        """Subtract two KSSingles"""
        result = copy(self)
        result.me = self.me - other.me
        result.mur = self.mur - other.mur
        result.muv = self.muv - other.muv
        result.magn = self.magn - other.magn
        return result

    def __rmul__(self, x):
        return self.__mul__(x)

    def __mul__(self, x):
        """Multiply a KSSingle with a number"""
        assert isinstance(x, (float, int))
        result = copy(self)
        result.me = self.me * x
        result.mur = self.mur * x
        result.muv = self.muv * x
        result.magn = self.magn * x
        return result

    def __truediv__(self, x):
        return self.__mul__(1. / x)

    __div__ = __truediv__

    def fromstring(self, string, dtype=float):
        l = string.split()
        self.i = int(l.pop(0))
        self.j = int(l.pop(0))
        self.pspin = int(l.pop(0))
        self.spin = int(l.pop(0))
        if dtype == float:
            self.k = 0
            self.weight = 1
        else:
            self.k = int(l.pop(0))
            self.weight = float(l.pop(0))
        self.energy = float(l.pop(0))
        self.fij = float(l.pop(0))
        self.mur = np.array([dtype(l.pop(0)) for i in range(3)])
        self.me = - self.mur * np.sqrt(self.energy * self.fij)
        self.muv = self.magn = None
        if len(l):
            self.muv = np.array([dtype(l.pop(0)) for i in range(3)])
        if len(l):
            self.magn = np.array([dtype(l.pop(0)) for i in range(3)])
        return None

    def outstring(self):
        if self.mur.dtype == float:
            string = '{0:d} {1:d}  {2:d} {3:d}  {4:.10g} {5:f}'.format(
                self.i, self.j, self.pspin, self.spin, self.energy, self.fij)
        else:
            string = (
                '{0:d} {1:d}  {2:d} {3:d} {4:d} {5:.10g}  {6:g} {7:g}'.format(
                    self.i, self.j, self.pspin, self.spin, self.k,
                    self.weight, self.energy, self.fij))
        string += '  '

        def format_me(me):
            string = ''
            if me.dtype == float:
                for m in me:
                    string += ' {0:.5e}'.format(m)
            else:
                for m in me:
                    string += ' {0.real:.5e}{0.imag:+.5e}j'.format(m)
            return string

        string += '  ' + format_me(self.mur)
        if self.muv is not None:
            string += '  ' + format_me(self.muv)
        if self.magn is not None:
            string += '  ' + format_me(self.magn)
        string += '\n'

        return string

    def __str__(self):
        string = '# <KSSingle> %d->%d %d(%d) eji=%g[eV]' % \
            (self.i, self.j, self.pspin, self.spin,
             self.energy * Hartree)
        if self.me.dtype == float:
            string += ' (%g,%g,%g)' % (self.me[0], self.me[1], self.me[2])
        else:
            string += ' kpt={0:d} w={1:g}'.format(self.k, self.weight)
            string += ' ('
            # use velocity form
            s = - np.sqrt(self.energy * self.fij)
            for c, m in enumerate(s * self.me):
                string += '{0.real:.5e}{0.imag:+.5e}j'.format(m)
                if c < 2:
                    string += ','
            string += ')'
        return string

    def __eq__(self, other):
        """KSSingles are considred equal when their indices are equal."""
        return (self.pspin == other.pspin and self.k == other.k and
                self.i == other.i and self.j == other.j)

    def __hash__(self):
        """Hash similar to __eq__"""
        if not hasattr(self, 'hash'):
            self.hash = hash((self.spin, self.k, self.i, self.j))
        return self.hash

    #
    # User interface: ##
    #

    def get_weight(self):
        return self.fij


def _get_and_distribute_wf(wfs, n, k, s):
    gd = wfs.gd
    wf = wfs.get_wave_function_array(n=n, k=k, s=s, realspace=True,
                                     periodic=False)
    if wfs.world.rank != 0:
        wf = gd.empty(dtype=wfs.dtype, global_array=True)
    wf = np.ascontiguousarray(wf)
    wfs.world.broadcast(wf, 0)
    wfd = gd.empty(dtype=wfs.dtype, global_array=False)
    wfd = gd.distribute(wf)
    return wfd



import numpy as np
import pytest

import gpaw.mpi as mpi
from gpaw import GPAW
from gpaw.test import equal
from gpaw.xas import XAS


@pytest.mark.later
def test_corehole_h2o(in_tmp_dir, add_cwd_to_setup_paths, gpw_files):
    # Generate setup for oxygen with half a core-hole:
    calc = GPAW(gpw_files['h2o_xas'])
    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e1_n = xas.eps_n
        de1 = e1_n[1] - e1_n[0]

    if mpi.size == 1:
        # calc = GPAW('h2o-xas.gpw')
        # poissonsolver=FDPoissonSolver(use_charge_center=True))
        # calc.initialize()
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e2_n = xas.eps_n
        w_n = np.sum(xas.sigma_cn.real**2, axis=0)
        de2 = e2_n[1] - e2_n[0]

        equal(de2, 2.064, 0.005)
        equal(w_n[1] / w_n[0], 2.22, 0.01)

        assert de1 == de2

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(x, y[0])
        plt.plot(x, sum(y))
        plt.show()



import pytest
from gpaw.response.g0w0 import G0W0
import numpy as np


@pytest.mark.response
def test_gw_anisotropic(in_tmp_dir, gpw_files, needs_ase_master):
    print(gpw_files)
    gw = G0W0(gpw_files['p4_pw'],
              'gw-test',
              nbands=15,
              ecut=30,
              eta=0.2,
              frequencies={'type': 'nonlinear', 'domega0': 0.3},
              truncation='2D',
              kpts=[(0.5, 0, 0), (0, 0, 0), (0, 0.5, 0)],
              bands=(9, 11))

    e_qp = gw.calculate()['qp']
    assert np.allclose(e_qp, [[[3.80581804, 12.04467414],
                               [3.21438974, 11.23852882],
                               [2.32178208, 11.6378957]]], atol=0.001)



import pytest
from ase import Atoms

from gpaw import GPAW
from gpaw.bfield import BField


@pytest.mark.later
@pytest.mark.serial
def test_b_field():
    """Hydrogen atom in a magnetic field."""
    L = 2.0
    atom = Atoms('H', magmoms=[1], cell=[L, L, L], pbc=True)
    atom.calc = GPAW(mode='pw')
    E1 = atom.get_potential_energy()
    a1, b1 = (atom.calc.get_eigenvalues(spin=s)[0] for s in [0, 1])

    B = 0.1

    # Collinear:
    atom.calc = GPAW(mode='pw',
                     external=BField((0, 0, B)))
    E2 = atom.get_potential_energy()
    a2, b2 = (atom.calc.get_eigenvalues(spin=s)[0] for s in [0, 1])

    assert E2 - E1 == pytest.approx(-B, abs=1e-6)
    assert a2 - a1 == pytest.approx(-B, abs=1e-6)
    assert b2 - b1 == pytest.approx(B, abs=1e-6)

    # Non-collinear:
    atom.calc = GPAW(mode='pw',
                     experimental={'magmoms': [(0.5, 0.5, 0)]},
                     external=BField([B, 0, 0]))
    E3 = atom.get_potential_energy()
    a3, b3 = atom.calc.get_eigenvalues()

    assert E3 - E1 == pytest.approx(-B, abs=2e-5)
    assert a3 - a1 == pytest.approx(-B, abs=3e-5)

    totmom_v, magmom_av = atom.calc.density.estimate_magnetic_moments()
    assert totmom_v == pytest.approx([1, 0, 0], abs=1e-5)



from pathlib import Path
from gpaw.lcao.analyse_basis import main


def test_analyse_basis(in_tmp_dir, capsys):
    argv = ['H.dzp.basis', '--save-figs']
    main(argv)
    captured = capsys.readouterr()
    out = captured.out
    assert '1s-sz' in out
    assert '1s-dz' in out
    assert 'p-type' in out
    assert Path('H.dzp.png').is_file()



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
Ref. to Kresse-paper ... XXX
"""

import numpy as np

from gpaw.utilities.blas import axpy
from gpaw.fd_operators import FDOperator


class BaseMixer:
    """Pulay density mixer."""

    def __init__(self, beta=0.1, nmaxold=3, weight=50.0, dtype=float):
        """Construct density-mixer object.

        Parameters
        ----------
        beta: float
            Mixing parameter between zero and one (one is most
            aggressive).
        nmaxold: int
            Maximum number of old densities.
        weight: float
            Weight parameter for special metric (for long wave-length
            changes).

        """

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight
        self.dtype = dtype

        self.dNt = None

        self.mix_rho = False

    def initialize_metric(self, gd):
        self.gd = gd

        if self.weight == 1:
            self.metric = None
        else:
            a = 0.125 * (self.weight + 7)
            b = 0.0625 * (self.weight - 1)
            c = 0.03125 * (self.weight - 1)
            d = 0.015625 * (self.weight - 1)
            self.metric = FDOperator(
                [a,
                 b, b, b, b, b, b,
                 c, c, c, c, c, c, c, c, c, c, c, c,
                 d, d, d, d, d, d, d, d],
                [(0, 0, 0),
                 (-1, 0, 0), (1, 0, 0),                  # b
                 (0, -1, 0), (0, 1, 0),                  # b
                 (0, 0, -1), (0, 0, 1),                  # b
                 (1, 1, 0), (1, 0, 1), (0, 1, 1),        # c
                 (1, -1, 0), (1, 0, -1), (0, 1, -1),     # c
                 (-1, 1, 0), (-1, 0, 1), (0, -1, 1),     # c
                 (-1, -1, 0), (-1, 0, -1), (0, -1, -1),  # c
                 (1, 1, 1), (1, 1, -1), (1, -1, 1),      # d
                 (-1, 1, 1), (1, -1, -1), (-1, -1, 1),   # d
                 (-1, 1, -1), (-1, -1, -1)               # d
                 ],
                gd, self.dtype).apply
            self.mR_G = gd.empty(dtype=self.dtype)

    def initialize(self, density):
        self.initialize_metric(density.gd)

    def reset(self):
        """Reset Density-history.

        Called at initialization and after each move of the atoms.

        my_nuclei:   All nuclei in local domain.
        """

        # History for Pulay mixing of densities:
        self.nt_iG = []  # Pseudo-electron densities
        self.R_iG = []  # Residuals
        self.A_ii = np.zeros((0, 0))
        self.dNt = None

        self.D_iap = []
        self.dD_iap = []

    def mix(self, nt_G, D_ap, phase_cd=None):
        iold = len(self.nt_iG)
        if iold > 0:
            if iold > self.nmaxold:
                # Throw away too old stuff:
                del self.nt_iG[0]
                del self.R_iG[0]
                del self.D_iap[0]
                del self.dD_iap[0]
                # for D_p, D_ip, dD_ip in self.D_a:
                #     del D_ip[0]
                #     del dD_ip[0]
                iold = self.nmaxold

            # Calculate new residual (difference between input and output)
            R_G = nt_G - self.nt_iG[-1]
            # Use np.absolute instead of np.fabs
            self.dNt = self.gd.integrate(np.absolute(R_G))
            self.R_iG.append(R_G)
            self.dD_iap.append([])
            for D_p, D_ip in zip(D_ap, self.D_iap[-1]):
                self.dD_iap[-1].append(D_p - D_ip)

            # Update matrix:
            A_ii = np.zeros((iold, iold))
            i1 = 0
            i2 = iold - 1

            if self.metric is None:
                mR_G = R_G
            else:
                mR_G = self.mR_G
                self.metric(R_G, mR_G, phase_cd=phase_cd)

            for R_1G in self.R_iG:
                # Inner product between new and old residues
                # XXX For now, use only real part of residues
                # For complex quantities a .conjugate should be added ??
                a = self.gd.comm.sum(np.vdot(R_1G.real, mR_G.real))
                if self.dtype == complex:
                    a += self.gd.comm.sum(np.vdot(R_1G.imag, mR_G.imag))

                A_ii[i1, i2] = a
                A_ii[i2, i1] = a
                i1 += 1
            A_ii[:i2, :i2] = self.A_ii[-i2:, -i2:]
            self.A_ii = A_ii

            try:
                B_ii = np.linalg.inv(A_ii)
            except np.linalg.LinAlgError:
                alpha_i = np.zeros(iold)
                alpha_i[-1] = 1.0
            else:
                alpha_i = B_ii.sum(1)
                try:
                    # Normalize:
                    alpha_i /= alpha_i.sum()
                except ZeroDivisionError:
                    alpha_i[:] = 0.0
                    alpha_i[-1] = 1.0

            # Calculate new input density:
            nt_G[:] = 0.0

            for D in D_ap:
                D[:] = 0.0
            beta = self.beta

            for i, alpha in enumerate(alpha_i):
                axpy(alpha, self.nt_iG[i], nt_G)
                axpy(alpha * beta, self.R_iG[i], nt_G)
                for D_p, D_ip, dD_ip in zip(D_ap, self.D_iap[i],
                                            self.dD_iap[i]):
                    axpy(alpha, D_ip, D_p)
                    axpy(alpha * beta, dD_ip, D_p)

        # Store new input density (and new atomic density matrices):
        self.nt_iG.append(nt_G.copy())
        self.D_iap.append([])
        for D_p in D_ap:
            self.D_iap[-1].append(D_p.copy())

    def estimate_memory(self, mem, gd):
        gridbytes = gd.bytecount()
        mem.subnode('nt_iG, R_iG', 2 * self.nmaxold * gridbytes)

    def __repr__(self):
        classname = self.__class__.__name__
        template = '%s(beta=%f, nmaxold=%d, weight=%f)'
        string = template % (classname, self.beta, self.nmaxold, self.weight)
        return string


class DummyMixer(BaseMixer):
    """Dummy mixer for TDDFT, i.e., it does not mix."""
    def mix(self, nt_G):
        pass

    def estimate_memory(self, mem, gd):
        pass


class Mixer(BaseMixer):
    """Mix spin up and down densities separately"""

    def initialize(self, density):
        self.mixers = []
        for s in range(density.nspins):
            mixer = BaseMixer(self.beta, self.nmaxold, self.weight)
            mixer.initialize_metric(density.gd)
            self.mixers.append(mixer)

    def mix(self, density):
        """Mix pseudo electron densities."""

        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()
        D_sap = []
        for s in range(density.nspins):
            D_sap.append([D_sp[s] for D_sp in D_asp])
        for nt_G, D_ap, mixer in zip(nt_sG, D_sap, self.mixers):
            mixer.mix(nt_G, D_ap)

    def reset(self):
        for mixer in self.mixers:
            mixer.reset()


class MixerSum(BaseMixer):
    """For pseudo electron densities, mix the total charge density and for
    density matrices, mix spin up and densities separately.
    Magnetization density is not mixed, i.e new magnetization density is used
    """

    def mix(self, density):
        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()

        # Mix density
        nt_G = density.nt_sG.sum(0)
        BaseMixer.mix(self, nt_G, D_asp)

        # Only new magnetization for spin density
        dnt_G = nt_sG[0] - nt_sG[1]
        # dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]

        # Construct new spin up/down densities
        nt_sG[0] = 0.5 * (nt_G + dnt_G)
        nt_sG[1] = 0.5 * (nt_G - dnt_G)


class MixerSum2(BaseMixer):
    """Mix the total pseudo electron density and the total density matrices.
    Magnetization density is not mixed, i.e new magnetization density is used.
    """

    def mix(self, density):

        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()

        # Mix density
        nt_G = density.nt_sG.sum(0)
        D_ap = [D_p[0] + D_p[1] for D_p in D_asp]
        BaseMixer.mix(self, nt_G, D_ap)

        # Only new magnetization for spin density
        dnt_G = nt_sG[0] - nt_sG[1]
        dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]

        # Construct new spin up/down densities
        nt_sG[0] = 0.5 * (nt_G + dnt_G)
        nt_sG[1] = 0.5 * (nt_G - dnt_G)
        for D_sp, D_p, dD_p in zip(D_asp, D_ap, dD_ap):
            D_sp[0] = 0.5 * (D_p + dD_p)
            D_sp[1] = 0.5 * (D_p - dD_p)


class MixerDif(BaseMixer):
    """Mix the charge density and magnetization density separately"""

    def __init__(self, beta=0.1, nmaxold=3, weight=50.0,
                 beta_m=0.7, nmaxold_m=2, weight_m=10.0):
        """Construct density-mixer object.

        Parameters
        ----------
        beta: float
            Mixing parameter between zero and one (one is most
            aggressive).
        nmaxold: int
            Maximum number of old densities.
        weight: float
            Weight parameter for special metric (for long wave-length
            changes).
        """

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

        self.beta_m = beta_m
        self.nmaxold_m = nmaxold_m
        self.weight_m = weight_m
        self.dNt = None

        self.mix_rho = False

    def initialize(self, density):
        assert density.nspins == 2
        self.mixer = BaseMixer(self.beta, self.nmaxold, self.weight)
        self.mixer.initialize_metric(density.gd)
        self.mixer_m = BaseMixer(self.beta_m, self.nmaxold_m, self.weight_m)
        self.mixer_m.initialize_metric(density.gd)

    def reset(self):
        self.mixer.reset()
        self.mixer_m.reset()

    def mix(self, density):

        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()

        # Mix density
        nt_G = density.nt_sG.sum(0)
        D_ap = [D_sp[0] + D_sp[1] for D_sp in D_asp]
        self.mixer.mix(nt_G, D_ap)

        # Mix magnetization
        dnt_G = nt_sG[0] - nt_sG[1]
        dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]
        self.mixer_m.mix(dnt_G, dD_ap)

        # Construct new spin up/down densities
        nt_sG[0] = 0.5 * (nt_G + dnt_G)
        nt_sG[1] = 0.5 * (nt_G - dnt_G)
        for D_sp, D_p, dD_p in zip(D_asp, D_ap, dD_ap):
            D_sp[0] = 0.5 * (D_p + dD_p)
            D_sp[1] = 0.5 * (D_p - dD_p)


class MixerRho(BaseMixer):
    def initialize(self, density):

        self.mix_rho = True
        self.initialize_metric(density.finegd)

    def mix(self, density):
        """Mix pseudo electron densities."""

        rhot_g = density.rhot_g
        BaseMixer.mix(self, rhot_g, [])


class MixerRho2(BaseMixer):
    def initialize(self, density):

        self.mix_rho = True
        self.initialize_metric(density.finegd)

    def mix(self, density):
        """Mix pseudo electron densities."""

        rhot_g = density.rhot_g
        BaseMixer.mix(self, rhot_g, density.D_asp.values())


class BaseMixer_Broydn:
    def __init__(self, beta=0.1, nmaxold=6):
        self.verbose = False
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = 1
        self.mix_rho = False

    def initialize(self, density):
        self.gd = density.gd

    def reset(self):
        self.step = 0
        self.d_nt_G = []
        self.d_D_ap = []
        self.nt_iG = []
        self.D_iap = []
        self.c_G = []
        self.v_G = []
        self.u_G = []
        self.u_D = []
        self.dNt = None

    def mix(self, nt_G, D_ap):
        if self.step > 2:
            del self.d_nt_G[0]
            for d_Dp in self.d_D_ap:
                del d_Dp[0]
        if self.step > 0:
            self.d_nt_G.append(nt_G - self.nt_iG[-1])
            for d_Dp, D_p, D_ip in zip(self.d_D_ap, D_ap, self.D_iap):
                d_Dp.append(D_p - D_ip[-1])
            fmin_G = self.gd.integrate(self.d_nt_G[-1] * self.d_nt_G[-1])
            self.dNt = self.gd.integrate(np.fabs(self.d_nt_G[-1]))
            if self.verbose:
                print('Mixer: broydn: fmin_G = %f fmin_D = %f' % fmin_G)
        if self.step == 0:
            self.eta_G = np.empty(nt_G.shape)
            self.eta_D = []
            for D_p in D_ap:
                self.eta_D.append(0)
                self.u_D.append([])
                self.D_iap.append([])
                self.d_D_ap.append([])
        else:
            if self.step >= 2:
                del self.c_G[:]
                if len(self.v_G) >= self.nmaxold:
                    del self.u_G[0]
                    del self.v_G[0]
                    for u_D in self.u_D:
                        del u_D[0]
                temp_nt_G = self.d_nt_G[1] - self.d_nt_G[0]
                self.v_G.append(temp_nt_G / self.gd.integrate(temp_nt_G
                                                              * temp_nt_G))
                if len(self.v_G) < self.nmaxold:
                    nstep = self.step - 1
                else:
                    nstep = self.nmaxold
                for i in range(nstep):
                    self.c_G.append(self.gd.integrate(self.v_G[i] *
                                                      self.d_nt_G[1]))
                self.u_G.append(self.beta * temp_nt_G +
                                self.nt_iG[1] - self.nt_iG[0])
                for d_Dp, u_D, D_ip in zip(self.d_D_ap, self.u_D, self.D_iap):
                    temp_D_ap = d_Dp[1] - d_Dp[0]
                    u_D.append(self.beta * temp_D_ap + D_ip[1] - D_ip[0])
                usize = len(self.u_G)
                for i in range(usize - 1):
                    a_G = self.gd.integrate(self.v_G[i] * temp_nt_G)
                    axpy(-a_G, self.u_G[i], self.u_G[usize - 1])
                    for u_D in self.u_D:
                        axpy(-a_G, u_D[i], u_D[usize - 1])
            self.eta_G = self.beta * self.d_nt_G[-1]
            for i, d_Dp in enumerate(self.d_D_ap):
                self.eta_D[i] = self.beta * d_Dp[-1]
            usize = len(self.u_G)
            for i in range(usize):
                axpy(-self.c_G[i], self.u_G[i], self.eta_G)
                for eta_D, u_D in zip(self.eta_D, self.u_D):
                    axpy(-self.c_G[i], u_D[i], eta_D)
            axpy(-1.0, self.d_nt_G[-1], nt_G)
            axpy(1.0, self.eta_G, nt_G)
            for D_p, d_Dp, eta_D in zip(D_ap, self.d_D_ap, self.eta_D):
                axpy(-1.0, d_Dp[-1], D_p)
                axpy(1.0, eta_D, D_p)
            if self.step >= 2:
                del self.nt_iG[0]
                for D_ip in self.D_iap:
                    del D_ip[0]
        self.nt_iG.append(np.copy(nt_G))
        for D_ip, D_p in zip(self.D_iap, D_ap):
            D_ip.append(np.copy(D_p))
        self.step += 1


class Mixer_Broydn(BaseMixer_Broydn):
    """Mix spin up and down densities separately"""

    def initialize(self, density):
        self.mixers = []
        for s in range(density.nspins):
            mixer = BaseMixer_Broydn()
            mixer.initialize(density)
            # mixer.initialize_metric(density.gd)
            self.mixers.append(mixer)

    def mix(self, density):
        """Mix pseudo electron densities."""

        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()
        D_sap = []
        for s in range(density.nspins):
            D_sap.append([D_sp[s] for D_sp in D_asp])
        for nt_G, D_ap, mixer in zip(nt_sG, D_sap, self.mixers):
            mixer.mix(nt_G, D_ap)

    def reset(self):
        for mixer in self.mixers:
            mixer.reset()


class MixerSum_Broydn(BaseMixer_Broydn):
    def mix(self, density):
        nt_sG = density.nt_sG
        D_asp = density.D_asp.values()

        # Mix density
        nt_G = density.nt_sG.sum(0)
        BaseMixer_Broydn.mix(self, nt_G, D_asp)

        # Only new magnetization for spin density
        dnt_G = nt_sG[0] - nt_sG[1]
        # dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]

        # Construct new spin up/down densities
        nt_sG[0] = 0.5 * (nt_G + dnt_G)
        nt_sG[1] = 0.5 * (nt_G - dnt_G)



import pytest
import numpy as np
from ase import Atoms
from ase.units import Bohr, Hartree
from gpaw.jellium import JelliumSlab
from gpaw import GPAW, Mixer
from gpaw.test import equal


@pytest.mark.later
@pytest.mark.libxc
def test_jellium(in_tmp_dir):
    rs = 5.0 * Bohr  # Wigner-Seitz radius
    h = 0.24          # grid-spacing
    a = 8 * h        # lattice constant
    v = 3 * a        # vacuum
    L = 8 * a       # thickness
    k = 6           # number of k-points (k*k*1)

    ne = a**2 * L / (4 * np.pi / 3 * rs**3)

    x = h / 4  # make sure surfaces are between grid-points
    bc = JelliumSlab(ne, z1=v - x, z2=v + L + x)

    surf = Atoms(pbc=(True, True, False),
                 cell=(a, a, v + L + v))
    surf.calc = GPAW(mode='fd',
                     background_charge=bc,
                     poissonsolver={'dipolelayer': 'xy'},
                     xc='LDA_X+LDA_C_WIGNER',
                     eigensolver='dav',
                     kpts=[k, k, 1],
                     h=h,
                     maxiter=300,
                     convergence={'density': 1e-5},
                     mixer=Mixer(0.3, 7, 100),
                     nbands=int(ne / 2) + 15,
                     txt='surface.txt')
    _ = surf.get_potential_energy()

    efermi = surf.calc.get_fermi_level()
    # Get (x-y-averaged) electrostatic potential
    # Must collect it from the CPUs
    # https://listserv.fysik.dtu.dk/pipermail/gpaw-users/2014-January/002524.html
    ham = surf.calc.hamiltonian
    v = (ham.finegd.collect(ham.vHt_g,
                            broadcast=True) * Hartree).mean(0).mean(0)

    # Get the work function
    phi1 = v[-1] - efermi

    equal(phi1, 2.715, 1e-3)
    # Reference value: Lang and Kohn, 1971, Theory of Metal Surfaces:
    # Work function
    # DOI 10.1103/PhysRevB.3.1215
    # r_s = 5, work function = 2.73 eV

    surf.calc.write('jellium.gpw')



import pytest

from gpaw.mpi import world
from gpaw.xc.fxc import FXCCorrelation


@pytest.mark.rpa
@pytest.mark.response
@pytest.mark.parametrize('gpw, kwargs, ref_energy, abstol', [
    ('h2_pw280_fulldiag', dict(xc='rALDA', nblocks=min(4, world.size)),
     -0.8509, 1e-3),
    ('h2_pw280_fulldiag', dict(xc='rAPBE'), -0.74555, 1e-3),
    ('h_pw280_fulldiag', dict(xc='rALDA'), 0.002757, 1e-4),
    ('h_pw280_fulldiag', dict(xc='rAPBE', nblocks=min(4, world.size)),
     0.01365, 1e-4),
])
def test_ralda_energy_H2(in_tmp_dir, gpw_files, scalapack, gpw,
                         kwargs,
                         ref_energy, abstol):
    gpw = gpw_files[gpw]
    fxc = FXCCorrelation(gpw, **kwargs, ecut=[200])

    energy = fxc.calculate()
    assert energy == pytest.approx(ref_energy, abs=abstol)



import sys
from optparse import OptionParser


def build_parser():
    description = 'Generate LCAO basis sets for the specified elements.'

    parser = OptionParser(usage='%prog [options] [elements]',
                          version='%prog 0.1', description=description)
    parser.add_option('-n', '--name', default=None, metavar='<name>',
                      help='name of generated basis files')
    parser.add_option('-t', '--type', default='dzp', metavar='<type>',
                      help='type of basis.  For example: sz, dzp, qztp, ' +
                      '4z3p.  [default: %default]')
    parser.add_option('-E', '--energy-shift', metavar='<energy>', type='float',
                      default=.1,
                      help='use given energy shift to determine cutoff')
    parser.add_option('-T', '--tail-norm', metavar='<norm>', type='string',
                      default='0.16,0.3,0.6', dest='tailnorm',
                      help='use the given fractions to define the split' +
                      '-valence cutoffs.  Default: [%default]')
    parser.add_option('-f', '--xcfunctional', default='PBE', metavar='<XC>',
                      help='Exchange-Correlation functional '
                      '[default: %default]')
    parser.add_option('-g', '--non-relativistic-guess', action='store_true',
                      help='Run non-scalar relativistic AE calculation for '
                      'initial guess')
    parser.add_option('--rcut-max', type='float', default=16.,
                      metavar='<rcut>',
                      help='max cutoff for confined atomic orbitals.  This ' +
                      'option has no effect on orbitals with smaller cutoff ' +
                      '[default/Bohr: %default]')
    parser.add_option('--rcut-pol-rel', type='float', default=1.0,
                      metavar='<rcut>',
                      help='polarization function cutoff relative to largest' +
                      ' single-zeta cutoff [default: %default]')
    parser.add_option('--rchar-pol-rel', type='float', default=None,
                      metavar='<rchar>',
                      help='characteristic radius of Gaussian when ' +
                      'not using interpolation scheme, relative to rcut')
    parser.add_option('--vconf-amplitude', type='float', default=12.,
                      metavar='<alpha>',
                      help='set proportionality constant of smooth '
                      'confinement potential [default: %default]')
    parser.add_option('--vconf-rstart-rel', type='float', default=.6,
                      metavar='<ri/rc>',
                      help='set inner cutoff for smooth confinement potential '
                      'relative to hard cutoff [default: %default]')
    parser.add_option('--vconf-sharp-confinement', action='store_true',
                      help='use sharp rather than smooth confinement '
                      'potential')
    parser.add_option('--debug', action='store_true',
                      help='use gpaw-DEBUG mode')
    parser.add_option('--lpol', type=int, default=None,
                      help='angular momentum quantum number '
                      'of polarization function.  '
                      'Default behaviour is to take the lowest l which is not '
                      'among the valence states.')
    parser.add_option('--jvalues',
                      help='explicitly specify which states to include.  '
                      'Numbering corresponds to generator\'s valence state '
                      'ordering.  '
                      'For example: 0,1,2.')
    parser.add_option('--save-setup', action='store_true',
                      help='save setup to file.')

    return parser


bad_density_warning = """\
Bad initial electron density guess!  Try rerunning the basis generator
with the '-g' parameter to run a separate non-scalar relativistic
all-electron calculation and use its resulting density as an initial
guess."""

very_bad_density_warning = """\
Could not generate non-scalar relativistic electron density guess,
or non-scalar relativistic guess was not good enough for the scalar
relativistic calculation.  You probably have to use the Python interface
to the basis generator in gpaw.atom.basis directly and choose very
smart parameters."""


def main():
    parser = build_parser()
    opts, args = parser.parse_args()
    # if __name__ == '__main__' and len(args) == 0:
    #     parser.print_help()
    #     sys.exit(0)

    from gpaw.atom.basis import BasisMaker
    from gpaw import ConvergenceError
    from gpaw.basis_data import parse_basis_name

    zetacount, polcount = parse_basis_name(opts.type)

    referencefiles = [None] * len(args)
    reference_atom_indices = [None] * len(args)
    if polcount > 0:
        symbols = []
        for i, arg in enumerate(args):
            # Parse argument as <symbol>:<reference-file>:<nuclear index>.
            symbol_and_file = arg.split(':')
            symbol = symbol_and_file[0]
            symbols.append(symbol)
            if len(symbol_and_file) > 1:
                referencefiles[i] = symbol_and_file[1]  # filename
            if len(symbol_and_file) == 3:
                reference_atom_indices[i] = int(symbol_and_file[2])
    else:
        symbols = args

    for symbol, referencefile, referenceindex in zip(symbols, referencefiles,
                                                     reference_atom_indices):
        try:
            bm = BasisMaker(symbol, name=opts.name, gtxt=None,
                            non_relativistic_guess=opts.non_relativistic_guess,
                            xc=opts.xcfunctional,
                            save_setup=opts.save_setup)
        except ConvergenceError:
            if opts.non_relativistic_guess:
                print(very_bad_density_warning, file=sys.stderr)
                import traceback
                traceback.print_exc()
            else:
                print(bad_density_warning, file=sys.stderr)
            continue

        tailnorm = [float(norm) for norm in opts.tailnorm.split(',')]
        vconf_args = None
        if not opts.vconf_sharp_confinement:
            vconf_args = opts.vconf_amplitude, opts.vconf_rstart_rel

        jvalues = None
        if opts.jvalues:
            jvalues = [int(j) for j in opts.jvalues.split(',')]

        basis = bm.generate(zetacount, polcount,
                            tailnorm=tailnorm,
                            energysplit=opts.energy_shift,
                            referencefile=referencefile,
                            referenceindex=referenceindex,
                            rcutpol_rel=opts.rcut_pol_rel,
                            rcutmax=opts.rcut_max,
                            rcharpol_rel=opts.rchar_pol_rel,
                            vconf_args=vconf_args,
                            l_pol=opts.lpol,
                            jvalues=jvalues)
        basis.write_xml()



"""Create softlinks to JTH datasets."""
import re
import sys
from pathlib import Path


if __name__ == '__main__':
    for path in Path(sys.argv[1]).glob('*.*-JTH.xml'):
        m = re.match(r'([A-Z][a-z]?)\.[A-Z]+_([A-Z]+)-JTH\.xml', path.name)
        assert m is not None
        symbol = m[1]
        xc = m[2]
        Path(f'{symbol}.jth.{xc}').symlink_to(path)



"""Basic test of elph/supercell/calculate_gradient

Check that calculate_gradient is capable of reading stuff properly.
"""
import ase.units as units
import numpy as np
import pytest
from ase.build import bulk
from gpaw import GPAW
from gpaw.elph import DisplacementRunner, Supercell
from gpaw.mpi import world


@pytest.mark.elph
def test_gradient(in_tmp_dir):
    # 2 atoms with one 1 valence electron each
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)
    assert len(atoms) == 2
    # Note: usually we need to disable point group symmetry for displacements,
    # but we won't run actual displacements, so it saves time not to bother.
    calc = GPAW(mode='lcao',
                basis='sz(dzp)',
                kpts={'size': (2, 2, 2), 'gamma': False},
                # symmetry={'point_group': False},
                txt='elph_displacements.txt')
    atoms.calc = calc

    elph = DisplacementRunner(atoms=atoms, calc=atoms.calc,
                              supercell=(1, 1, 1), name='elph',
                              calculate_forces=False)
    elph.indices = []
    elph.run()

    # Barrier will be included in elph.run() in the future:
    # https://gitlab.com/ase/ase/-/merge_requests/2903
    world.barrier()

    Vt_sG = elph.cache['eq']['Vt_sG']
    dH_all_asp = elph.cache['eq']['dH_all_asp']

    # create displaced entries without calculation to save time
    V1expected = []
    for a in (0, 1):
        for i in range(3):
            V1expected.append(2 * 0.1 * (a + 1) * (i + 1) /
                              (2 * 0.01 / units.Bohr))
            for sign in [-1, 1]:
                disp = elph._disp(a, i, sign)
                with elph.cache.lock(disp.name) as handle:
                    if handle is None:
                        continue
                    try:
                        Vfake = Vt_sG + sign * 0.1 * (a + 1) * (i + 1)
                        # dHfake = {}
                        # for atom in dH_all_asp.keys():
                        #    dHfake[atom] = dH_all_asp[atom] + (sign * 0.2 *
                        #                                       (a+1) * (i+1))
                        result = {'Vt_sG': Vfake, 'dH_all_asp': dH_all_asp}
                        handle.save(result)
                    finally:
                        pass

    # calculate gradient
    V1t_xsG, dH1_xasp = Supercell.calculate_gradient('elph')
    assert np.array(V1expected) == pytest.approx(V1t_xsG[:, 0, 0, 0, 0])
    # should probably add something about dH as well, but the principle is the
    # same



import re
import numpy as np

from ase.utils import IOContext

from gpaw.lcaotddft.observer import TDDFTObserver


def calculate_quadrupole_moments(gd, rho_g, center_v):
    # Center relative to the cell center
    # center_v += 0.5 * gd.cell_cv.sum(0)
    r_vg = gd.get_grid_point_coordinates()
    qm_i = np.zeros(6, dtype=float)
    i = 0
    for v1 in range(3):
        x1_g = r_vg[v1] - center_v[v1]
        for v2 in range(v1, 3):
            x2_g = r_vg[v2] - center_v[v2]
            qm_i[i] = -gd.integrate(x1_g * x2_g * rho_g)
            i += 1
    return qm_i


class QuadrupoleMomentWriter(TDDFTObserver):
    version = 1

    def __init__(self, paw, filename, center=[0, 0, 0], density='comp',
                 interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        self.ioctx = IOContext()
        if paw.niter == 0:
            # Initialize
            if isinstance(center, str) and center == 'center':
                self.center_v = 0.5 * paw.density.finegd.cell_cv.sum(0)
            else:
                assert len(center) == 3
                self.center_v = center
            self.density_type = density
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='w')
        else:
            # Read and continue
            self.read_header(filename)
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='a')

    def _write(self, line):
        self.fd.write(line)
        self.fd.flush()

    def _write_header(self, paw):
        if paw.niter != 0:
            return
        line = '# %s[version=%s]' % (self.__class__.__name__, self.version)
        line += '(center=[%.6f, %.6f, %.6f]' % tuple(self.center_v)
        line += ', density=%s)\n' % repr(self.density_type)
        line += ('# %15s %15s %22s %22s %22s %22s %22s %22s\n' %
                 ('time', 'norm', 'x^2', 'xy', 'xz', 'y^2', 'yz', 'z^2'))
        self._write(line)

    def read_header(self, filename):
        with open(filename, 'r') as f:
            line = f.readline()
        regexp = (r"^(?P<name>\w+)\[version=(?P<version>\d+)\]\(center=\["
                  r"(?P<center0>[-+0-9\.]+), "
                  r"(?P<center1>[-+0-9\.]+), "
                  r"(?P<center2>[-+0-9\.]+)\], "
                  r"density='(?P<density>\w+)'\)$")
        m = re.match(regexp, line[2:])
        assert m is not None, 'Unknown fileformat'
        assert m.group('name') == self.__class__.__name__
        assert int(m.group('version')) == self.version
        self.center_v = [float(m.group('center%d' % v)) for v in range(3)]
        self.density_type = m.group('density')

    def _write_kick(self, paw):
        line = '# Kick: %s' % paw.kick_ext
        line += '; Time = %.8lf' % paw.time
        line += '\n'
        self._write(line)

    def _write_dm(self, paw):
        time = paw.time
        density = paw.density
        if self.density_type == 'comp':
            rho_g = density.rhot_g
            gd = density.finegd
        elif self.density_type == 'pseudo':
            rho_g = density.nt_sg.sum(axis=0)
            gd = density.finegd
        elif self.density_type == 'pseudocoarse':
            rho_g = density.nt_sG.sum(axis=0)
            gd = density.gd
        else:
            raise RuntimeError('Unknown density type: %s' % self.density_type)

        norm = gd.integrate(rho_g)
        qm = calculate_quadrupole_moments(gd, rho_g, center_v=self.center_v)
        line = (('%20.8lf %20.8le' + ' %22.12le' * len(qm) + '\n') %
                ((time, norm) + tuple(qm)))
        self._write(line)

    def _update(self, paw):
        if paw.action == 'init':
            self._write_header(paw)
        elif paw.action == 'kick':
            self._write_kick(paw)
        self._write_dm(paw)

    def __del__(self):
        self.ioctx.close()



import _gpaw
from gpaw.xc.kernel import XCKernel
from gpaw import debug

short_names = {
    'LDA': 'LDA_X+LDA_C_PW',
    'PW91': 'GGA_X_PW91+GGA_C_PW91',
    'PBE': 'GGA_X_PBE+GGA_C_PBE',
    'PBEsol': 'GGA_X_PBE_SOL+GGA_C_PBE_SOL',
    'revPBE': 'GGA_X_PBE_R+GGA_C_PBE',
    'RPBE': 'GGA_X_RPBE+GGA_C_PBE',
    'BLYP': 'GGA_X_B88+GGA_C_LYP',
    'HCTH407': 'GGA_XC_HCTH_407',
    'WC': 'GGA_X_WC+GGA_C_PBE',
    'AM05': 'GGA_X_AM05+GGA_C_AM05',
    'M06-L': 'MGGA_X_M06_L+MGGA_C_M06_L',
    'TPSS': 'MGGA_X_TPSS+MGGA_C_TPSS',
    'revTPSS': 'MGGA_X_REVTPSS+MGGA_C_REVTPSS',
    'mBEEF': 'MGGA_X_MBEEF+GGA_C_PBE_SOL',
    'SCAN': 'MGGA_X_SCAN+MGGA_C_SCAN'}


class LibXC(XCKernel):
    """Functionals from libxc."""
    def __init__(self, name):
        if not hasattr(_gpaw, 'lxcXCFuncNum'):
            raise NameError(
                f'Unable to use {name}: GPAW not compiled with LibXC!')
        self.name = name
        self.omega = None
        self.initialize(nspins=1)

    def initialize(self, nspins):
        self.nspins = nspins
        name = short_names.get(self.name, self.name)
        number = _gpaw.lxcXCFuncNum(name)
        if number is not None:
            f = number
            xc = -1
            x = -1
            c = -1
            if '_XC_' in name:
                xc = f
            elif '_C_' in name:
                c = f
            else:
                x = f
        else:
            try:
                x, c = name.split('+')
            except ValueError:
                raise NameError('Unknown functional: "%s".' % name)
            xc = -1
            x = _gpaw.lxcXCFuncNum(x)
            c = _gpaw.lxcXCFuncNum(c)
            if x is None or c is None:
                raise NameError('Unknown functional: "%s".' % name)

        self.xc = _gpaw.lxcXCFunctional(xc, x, c, nspins)
        self.set_omega()

        if self.xc.is_mgga():
            self.type = 'MGGA'
        elif self.xc.is_gga():
            self.type = 'GGA'
        else:
            self.type = 'LDA'

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg=None, dedsigma_xg=None,
                  tau_sg=None, dedtau_sg=None):
        if debug:
            self.check_arguments(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                                 tau_sg, dedtau_sg)
        nspins = len(n_sg)
        if self.nspins != nspins:
            self.initialize(nspins)

        self.xc.calculate(e_g.ravel(), n_sg, dedn_sg,
                          sigma_xg, dedsigma_xg,
                          tau_sg, dedtau_sg)

    def set_omega(self, omega=None):
        """Set the value of gamma/omega in RSF."""
        if omega is not None:
            self.omega = omega
        if self.omega is not None:
            if not self.xc.set_omega(self.omega):
                raise ValueError('Tried setting omega on non RSF hybrid.')



from __future__ import annotations

from pathlib import Path
from typing import IO, Any, Union

import ase.io.ulm as ulm
import gpaw
import gpaw.mpi as mpi
import numpy as np
from ase.io.trajectory import read_atoms, write_atoms
from ase.units import Bohr, Ha
from gpaw.core.atom_arrays import AtomArraysLayout
from gpaw.new.builder import builder as create_builder
from gpaw.new.calculation import DFTCalculation, DFTState, units
from gpaw.new.density import Density
from gpaw.new.input_parameters import InputParameters
from gpaw.new.logger import Logger
from gpaw.new.potential import Potential
from gpaw.utilities import unpack, unpack2
from gpaw.typing import DTypeLike

ENERGY_NAMES = ['kinetic', 'coulomb', 'zero', 'external', 'xc', 'entropy',
                'total_free', 'total_extrapolated',
                'band', 'stress']


def write_gpw(filename: str,
              atoms,
              params,
              calculation: DFTCalculation,
              skip_wfs: bool = True) -> None:

    comm = calculation.comm

    writer: ulm.Writer | ulm.DummyWriter
    if comm.rank == 0:
        writer = ulm.Writer(filename, tag='gpaw')
    else:
        writer = ulm.DummyWriter()

    with writer:
        writer.write(version=4,
                     gpaw_version=gpaw.__version__,
                     ha=Ha,
                     bohr=Bohr)

        write_atoms(writer.child('atoms'), atoms)

        # Note that 'non_collinear_magmoms' is not an ASE standard name!
        results = {key: value * units[key]
                   for key, value in calculation.results.items()
                   if key != 'non_collinear_magmoms'}
        writer.child('results').write(**results)

        p = {k: v for k, v in params.items() if k not in ['txt', 'parallel']}
        # ULM does not know about numpy dtypes:
        if 'dtype' in p:
            p['dtype'] = np.dtype(p['dtype']).name
        writer.child('parameters').write(**p)

        state = calculation.state
        state.density.write(writer.child('density'))
        state.potential._write_gpw(writer.child('hamiltonian'),
                                   calculation.state.ibzwfs)
        wf_writer = writer.child('wave_functions')
        state.ibzwfs.write(wf_writer, skip_wfs)

        if not skip_wfs and params.mode['name'] == 'pw':
            write_wave_function_indices(wf_writer,
                                        state.ibzwfs,
                                        state.density.nt_sR.desc)

    comm.barrier()


def write_wave_function_indices(writer, ibzwfs, grid):
    if ibzwfs.band_comm.rank != 0:
        return
    if ibzwfs.domain_comm.rank != 0:
        return

    kpt_comm = ibzwfs.kpt_comm
    ibz = ibzwfs.ibz
    nG = ibzwfs.get_max_shape(global_shape=True)[-1]

    writer.add_array('indices', (len(ibz), nG), np.int32)

    index_G = np.zeros(nG, np.int32)
    size = tuple(grid.size)
    if ibzwfs.dtype == float:
        size = (size[0], size[1], size[2] // 2 + 1)

    for k, rank in enumerate(ibzwfs.rank_k):
        if rank == kpt_comm.rank:
            wfs = ibzwfs.wfs_qs[ibzwfs.q_k[k]][0]
            i_G = wfs.psit_nX.desc.indices(size)
            index_G[:len(i_G)] = i_G
            index_G[len(i_G):] = -1
            if rank == 0:
                writer.fill(index_G)
            else:
                kpt_comm.send(index_G, 0)
        elif kpt_comm.rank == 0:
            kpt_comm.receive(index_G, rank)
            writer.fill(index_G)


def read_gpw(filename: Union[str, Path, IO[str]],
             *,
             log: Union[Logger, str, Path, IO[str]] = None,
             comm=None,
             parallel: dict[str, Any] = None,
             dtype: DTypeLike = None):
    """
    Read gpw file

    Returns
    -------
    atoms, calculation, params, builder
    """
    parallel = parallel or {}

    if not isinstance(log, Logger):
        log = Logger(log, comm or mpi.world)

    comm = log.comm

    log(f'Reading from {filename}')

    reader = ulm.Reader(filename)
    bohr = reader.bohr
    ha = reader.ha

    atoms = read_atoms(reader.atoms)
    kwargs = reader.parameters.asdict()
    kwargs['parallel'] = parallel

    if 'dtype' in kwargs:
        kwargs['dtype'] = np.dtype(kwargs['dtype'])

    # kwargs['nbands'] = reader.wave_functions.eigenvalues.shape[-1]

    for old_keyword in ['fixdensity', 'txt']:
        kwargs.pop(old_keyword, None)

    params = InputParameters(kwargs, warn=False)
    builder = create_builder(atoms, params, comm)

    if comm.rank == 0:
        nt_sR_array = reader.density.density * bohr**3
        vt_sR_array = reader.hamiltonian.potential / ha
        if builder.xc.type == 'MGGA':
            taut_sR_array = reader.density.ked * (bohr**3 / ha)
            dedtaut_sR_array = reader.hamiltonian.mgga_potential * bohr**-3
        D_sap_array = reader.density.atomic_density_matrices
        dH_sap_array = reader.hamiltonian.atomic_hamiltonian_matrices / ha
        shape = nt_sR_array.shape[1:]
    else:
        nt_sR_array = None
        vt_sR_array = None
        taut_sR = None
        dedtaut_sR = None
        D_sap_array = None
        dH_sap_array = None
        shape = None

    if builder.grid.global_shape() != mpi.broadcast(shape, comm=comm):
        # old gpw-file:
        kwargs.pop('h', None)
        kwargs['gpts'] = nt_sR_array.shape[1:]
        params = InputParameters(kwargs, warn=False)
        builder = create_builder(atoms, params, comm)

    if dtype is not None:
        params.mode['dtype'] = dtype

    (kpt_comm, band_comm, domain_comm, kpt_band_comm) = (
        builder.communicators[x] for x in 'kbdD')

    nt_sR = builder.grid.empty(builder.ncomponents)
    vt_sR = builder.grid.empty(builder.ncomponents)

    if builder.xc.type == 'MGGA':
        taut_sR = builder.grid.empty(builder.ncomponents)
        dedtaut_sR = builder.grid.empty(builder.ncomponents)
    else:
        taut_sR = None
        dedtaut_sR = None

    atom_array_layout = AtomArraysLayout([(setup.ni * (setup.ni + 1) // 2)
                                          for setup in builder.setups],
                                         atomdist=builder.atomdist)
    D_asp = atom_array_layout.empty(builder.ncomponents)
    dtype = float if builder.ncomponents < 4 else complex
    dH_asp = atom_array_layout.new(dtype=dtype).empty(builder.ncomponents)

    if kpt_band_comm.rank == 0:
        nt_sR.scatter_from(nt_sR_array)
        vt_sR.scatter_from(vt_sR_array)
        if builder.xc.type == 'MGGA':
            taut_sR.scatter_from(taut_sR_array)
            dedtaut_sR.scatter_from(dedtaut_sR_array)
        D_asp.scatter_from(D_sap_array)
        dH_asp.scatter_from(dH_sap_array)
    if reader.version < 4:
        convert_to_new_packing_convention(D_asp, density=True)
        convert_to_new_packing_convention(dH_asp)

    kpt_band_comm.broadcast(nt_sR.data, 0)
    kpt_band_comm.broadcast(vt_sR.data, 0)
    if builder.xc.type == 'MGGA':
        kpt_band_comm.broadcast(taut_sR.data, 0)
        kpt_band_comm.broadcast(dedtaut_sR.data, 0)
    kpt_band_comm.broadcast(D_asp.data, 0)
    kpt_band_comm.broadcast(dH_asp.data, 0)

    if reader.version >= 4:
        if comm.rank == 0:
            vHt_x_array = reader.hamiltonian.electrostatic_potential / ha
        else:
            vHt_x_array = None
        vHt_x = builder.electrostatic_potential_desc.empty()
        if kpt_band_comm.rank == 0:
            vHt_x.scatter_from(vHt_x_array)
        kpt_band_comm.broadcast(vHt_x.data, 0)
    else:
        vHt_x = None

    density = Density.from_data_and_setups(
        nt_sR, taut_sR, D_asp.to_full(),
        builder.params.charge,
        builder.setups,
        builder.get_pseudo_core_densities(),
        builder.get_pseudo_core_ked())
    energies = {name: reader.hamiltonian.get(f'e_{name}', np.nan) / ha
                for name in ENERGY_NAMES}
    penergies = {key: e for key, e in energies.items()
                 if not key.startswith('total')}
    e_band = penergies.pop('band', np.nan)
    e_entropy = penergies.pop('entropy')
    penergies['kinetic'] -= e_band

    potential = Potential(vt_sR, dH_asp.to_full(), dedtaut_sR, penergies,
                          vHt_x)

    ibzwfs = builder.read_ibz_wave_functions(reader)
    ibzwfs.energies = {
        'band': e_band,
        'entropy': e_entropy,
        'extrapolation': (energies['total_extrapolated'] -
                          energies['total_free'])}

    calculation = DFTCalculation(
        DFTState(ibzwfs, density, potential),
        builder.setups,
        builder.create_scf_loop(),
        pot_calc=builder.create_potential_calculator(),
        log=log)

    results = {key: value / units[key]
               for key, value in reader.results.asdict().items()}

    if results:
        log(f'Read {", ".join(sorted(results))}')

    calculation.results = results

    if builder.mode in ['pw', 'fd']:  # fd = finite difference
        data = ibzwfs.wfs_qs[0][0].psit_nX.data
        if not hasattr(data, 'fd'):  # fd = file descriptor
            reader.close()
    else:
        reader.close()

    return atoms, calculation, params, builder


def convert_to_new_packing_convention(a_asp, density=False):
    """Convert from old to new convention.

    ::

        1 2 3      1 2 4
        . 4 5  ->  . 3 5
        . . 6      . . 6
    """
    for a_sp in a_asp.values():
        if density:
            a_sii = unpack2(a_sp)
        else:
            a_sii = unpack(a_sp)
        L = np.tril_indices(a_sii.shape[1])
        a_sp[:] = a_sii[(...,) + L]
